#!/usr/local/bin/perl 
#
# File: qmail-scanner-queue.pl
# Version: 2.01
#
# Author: Jason L. Haar <jhaar@users.sourceforge.net>
# 
# This file was auto-generated by:
#
# ./configure --spooldir /var/spool/qscand --qmaildir /var/qmail --bindir /usr/local/bin --qmail-queue-binary /var/qmail/bin/qmail-queue --admin postmaster --domain gohan11.freebsd.org --admin-description "System Anti-Virus Administrator" --notify psender,precips --local-domains gohan11.freebsd.org --max-scan-size 100000000 --silent-viruses auto --sa-quarantine 0 --lang en_GB --debug 1 --unzip 0 --max-zip-size 1000000000 --add-dscr-hdrs 0 --normalize no --archive 0 --redundant yes --skip-text-msgs 1 --log-details syslog --log-crypto 0 --fix-mime 0  --ignore-eol-check 0 --scanners "auto" --no-QQ-check 1
# 
# Description: This is a replacement/add-on for Qmail 1.0.3's qmail-queue.
# It can call several blocking programs - such as virus scanners - on every 
# SMTP-received Email message, checking for viruses and blocked filenames, 
# only allowing the message to continue if it passes the tests.
#
#   Copyright (C) 1999,2000,2001 the people mentioned above
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 1, or (at your option)
#   any later version. See <URL:http://www.gnu.org/copyleft/gpl.html>
#   for a copy.
# 
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   The software is provided as is. Please bear in mind that we have 
#   done this in my spare time. While it is as accurate as we could 
#   make it there is a reasonable chance that there are mistakes 
#   somewhere in here. If you email me and tell me about them, I will 
#   be happy to fix them but I can't take responsibility for your system. 
#   Basically use this at your own risk. 
#
#####################################################################

#####################################################################
##
## Required Packages
##
## Qmail-1.03
## Perl 5.005_03+
## Maildrop-0.73
## Bruce Guenter's QMAILQUEUE patch <URL:http://www.qmail.org/qmailqueue-patch>
## Perl module Time::HiRes and DB_File
##
##
## So-far tested Virus scanners:
##     Trend's Virus scanner for Linux
##     MacAfee's (NAI's) virus scanner for Linux
##     Sophos's virus scanner for Linux
##     H+BEDV's antivir scanner for Linux
##     F-Secure's fsav scanner for Linux
##     P-Prot for Linux
##     Sophie (daemonized Sophos scanner)
##     Trophie (daemonized Trend scanner)
##     ...and more - see README for full list
## 
#####################################################################

#####################################################################
##
## Site-specific config
##
#####################################################################


delete @ENV{qw(IFS CDPATH ENV BASH_ENV QMAILMFTFILE QMAILINJECT)};

use strict 'vars', 'subs';

#Set locale to "C" (English). That way any string checks on forked apps 
#will tend to be in English - simplifying/standardizing regex matches
my $orig_locale=$ENV{'LC_ALL'};
$ENV{'LC_ALL'}= $ENV{'LANG'} = $ENV{'LANGUAGE'} = 'C';
POSIX::setlocale(&POSIX::LC_ALL,'C');

use Sys::Syslog qw(:DEFAULT setlogsock);
setlogsock('unix');

my $VERSION="2.01";

#Mail header to add to each scanned message to report stuff in...
#Default is to not generate them ($descriptive_hdrs = 0) - as  that
#info is also in the Received: headers...
my $descriptive_hdrs=0;
my $V_HEADER="X-Qmail-Scanner";
my($qsmsgid);
$qsmsgid=tolower("$V_HEADER-message-id");

my($qscan_account)='qscand';

#From: line  information used when making reports
my $V_FROM='postmaster@gohan11.freebsd.org';
my $V_FROMNAME='System Anti-Virus Administrator';

# Address carbon-copied on any virus reports
my $QUARANTINE_CC='postmaster@gohan11.freebsd.org';

#Array of local domains that are checked against for
#deciding whether or not to send recipient alerts to
my @local_domains_array=('gohan11.freebsd.org');

# Array of virus that we don't want to inform the sender of.
my @silent_viruses_array=('klez','bugbear','hybris','yaha','braid','nimda','tanatos','sobig','winevar','palyh','fizzer','gibe','cailont','lovelorn','swen','dumaru','sober','hawawi','holar-i','mimail','poffer','bagle','worm.galil','mydoom','worm.sco','tanx','novarg','@mm');


#Array of virus scanners used must point to subroutines
my @scanner_array=();

#Array of virtual headers used within perlscanner 
my @virtualheaders_array=("MAILFROM","RCPTTO","REMOTEIPADDR","ZIPPASSWORDPROTECTED","ISSENSITIVEANDNOCRYPTO","CRYPTODETAILS","FILELENGTHTOOLONG","FILEDOUBLEBARRELED","FILECLSID");

#Addresses that should be alerted of any quarantined Email
my $NOTIFY_ADDRS='psender,precips';

#Try to fix bad MIME messages before passing to MIME unpacker
my $BAD_MIME_CHECKS='0';

#Block password protected zip files
#my $BLOCK_PASSWORD_PROTECTED_ARCHIVES='QUARANTINE_PASSWORD_PROTECTED';

#Disable just the EOL char check instead of all of BAD_MIME_CHECKS
my $IGNORE_EOL_CHECK='0';

# The full path to qmail programs we'll need.
my $qmailinject = '/var/qmail/bin/qmail-inject';
my $qmailqueue  = '/var/qmail/bin/qmail-queue';

# What directory to use for storing temporary files.
my $scandir = '/var/spool/qscand';

#Where the Q-S configs are
my $configdir = '/var/spool/qscand';

#Where the Q-S logs live
my $logdir = '/var/spool/qscand';

#What maildir folder to store working files in
my $wmaildir='working';

#What maildir folder to store virus-infected msgs in
my $vmaildir='viruses';

#What maildir folder to store policy-blocked msgs in
my $pmaildir='policy';

#What maildir folder to store high-scoring SPAM in (instead of passing it on)
#NOTE: this only gets used if 0 set 
my $smaildir='spam';

#What maildir folder to archive received Email in instead of deleting
my $archiveit='0';
my $archivedir='archives';

#Name of file in $scandir where debugging output goes
my $debuglog="qmail-queue.log";

#Name of file where quarantine reports go (for long-term storage)
my $quarantinelog="quarantine.log";

#Generate nice random filename
my ($sysname, $hostname, $release, $version, $machine) = uname();
#my $hostname='gohan11.freebsd.org'; #could get via call I suppose...

#If you trust the virus scanners handling of mbox format itself
#you may want to let it have a go at the "raw" message, and original
#zip files if present
my $redundant_scanning='1';

#If you want to log via file/syslog information of all Email
# that passes through your system (from/to/subj/size/attachments)
my $log_details="syslog";

#If you'd like Q-S to report which messages are PGP or S/MIME,
#turn this on
my $log_crypto="0";

#Max size of message allowed to be scanned - 100Mbytes by default 
#DO NOT SET LOWER THAN 10Mbytes!!!!!
my $MAX_SCAN_SIZE=100000000;

#bypass all AV/Spam scanning - but still do perlscan checks
my $SKIP_SCANNING=0;

#Full path to file in which virus-scanner versioning info is kept
my $versionfile="$logdir/qmail-scanner-queue-version.txt";

#DB file (without extension) where bad filenames are kept.
# You edit $db_filename.txt, and "qmail-scanner-queue.pl -g" generates $db_filename.db
my $db_filename="$configdir/quarantine-events";

my $MAX_NUM_UNPACKED_FILES=10000; #10,000 is stupidly high. This rule exists but is never 
                                  #expected to trigger normally

#What locale is used on this system
#$sys_locale="LOCALE";

#Full paths to binaries used within this script follow - small performance
#improvement :-)


my $mimeunpacker_binary='/usr/local/bin/reformime ';
my $unzip_binary='/usr/local/bin/unzip';
my $unzip_options='-Pxx5531xx';
my $max_zip_size='1000000000';
my $tnef_binary='/usr/local/bin/tnef';
my $rm_binary='/bin/rm';
my $grep_binary='/usr/bin/grep';
my $find_binary='/usr/bin/find';
my $uudecode_binary='/usr/bin/uudecode';
my $uudecode_pipe='-o -';


my $uvscan_binary='';
my $csav_binary='';
my $nod32_binary='';
my $nod32upd_binary='';
my $sweep_binary='';
my $sophie_binary='';
my $trophie_binary='';
my $iscan_binary='';
my $hbedv_binary='';
my $hbedv_options='';
my $avp_binary='';
my $avpdaemon_binary='';
my $fprot_binary='';
my $fsecure_binary='';
my $inocucmd_binary='';
my $ravlin_binary='';
my $vexira_binary='';
my $bitdefender_binary='';
my $clamscan_binary='';
my $clamscan_options="-r -m --unzip --unrar --unzoo --lha --disable-summary --max-recursion=10 --max-space=100000";
my $clamdscan_binary='';
my $clamdscan_options="--no-summary";
my $spamc_binary='';
my $spamc_options='';
my $spamc_subject='';
my $spamassassin_binary='';
my $sa_quarantine_over='0';
my ($sa_comment,$sa_level);
my $sa_symbol='+';
my ($tag_score)="";
my $SNEAKY_WINDOWS_EXTENSIONS="exe|w[pm][szd]|vcf|nws|cmd|bat|pif|sc[rt]|dll|ocx|do[ct]|xl[swt]|p[po]t|pps|vb[se]?|hta|p[lm]|sh[bs]|hlp|chm|eml|ws[cfh]|ad[ep]|jse?|md[abew]|ms[ip]|reg|as[dfx]|cil|cpl";
my $VALID_WINDOWS_EXTENSIONS="rtf|pdf|sav|htm|html|pst|ost|txt|gif|jpeg|mpeg|jpg|png|mny|wav|tif|$SNEAKY_WINDOWS_EXTENSIONS";
my $passwd_protected_zip;

#Little workaround. Apparently people send docs of the form "my.domainname.com.doc" - so you can get false positives
#due to ".com". So don't add  ".com" to SNEAKY_WINDOWS_EXTENSIONS until after VALID_WINDOWS_EXTENSIONS is defined
#So now "file.com.tif" won't trigger, but file.tif.com will.
$SNEAKY_WINDOWS_EXTENSIONS="$SNEAKY_WINDOWS_EXTENSIONS|com";


$ENV{'PATH'}='/bin:/usr/bin';

my $SCANINFO='';

my $MAX_FILE_LENGTH=100;
my $MAX_NUM_HDRS=140;
my $QE_LEN=20;

#Maximum amount of time we allow Q-S to run before returning
# a temp failure. This is so remote SMTP servers don't get confused
# over whether or not they have delivered to a SMTP server
# that's refused to say "OK" for over an hour...
# We'll default to 20 minutes. If the scanner loop takes more than 20 
# minutes to scan the message, then something *must* be wrong with the
# scanner. 
my $MAXTIME=20*60;

#Finally, are you sure your virus scanners can unpack zip files? 
#Turn this on to force Qmail-Scanner to unzip for you
my $force_unzip=0;

#Descriptive string to use in generated Email
my $destring="virus";

#####################################################################
## 
## End of site-specific settings
##
#####################################################################



#Want debugging? Enable this and read $logdir/qmail-queue.log
my $DEBUG='1';

my @uufile_list = ();
my @attachment_list = ();
my @zipfile_list = ();

#Want microsec times for debugging
use Time::HiRes qw( usleep ualarm gettimeofday tv_interval );
use POSIX;

use vars qw/ $opt_v $opt_h $opt_g $opt_r $opt_z/;

use Getopt::Std;

#my ($opt_v,$opt_h,$opt_g,$opt_r,$opt_z);

getopts('vhgrz');

my ($start_time,$last_time);
$start_time = $last_time = [gettimeofday];

(my $prog=$0) =~ s/^.*\///g;

if ( $opt_h ) {
  print "

$prog

    -h - This help
    -v - show details about this install. 
         Please include in any bug reports.
    -z - gather virus scanner/DAT versions 
         and cleanup old temp files
    -g - generate perlscanner database
    -r - read from perlscanner database\n";
  exit;
}


if ( $opt_g || $opt_r) {
  &generate_quarantine_db;
    exit 0;
}

if ( $opt_v ) {
  &show_version;
  exit 0;
}

chdir($scandir);
umask(0007);

if (! -d "$scandir/tmp") {
  mkdir("$scandir/tmp",0750) || &error_condition("cannot create $scandir/tmp - $!");
}

my ($quarantine_event,$quarantine_event_tmp)=0;
my ($quarantine_DOS,$quarantine_spam)=0;

my $file_id = &uniq_id();

#For security reasons, tighten the follow vars...
$ENV{'SHELL'} = '/bin/sh' if exists $ENV{SHELL};
$ENV{'TMP'} = $ENV{'TMPDIR'} = "$scandir/tmp/$file_id";
#$ENV{'QMAILSUSER'} = $ENV{'QMAILSHOST'} = '';



if ($mimeunpacker_binary =~ /reformime/) {
  $mimeunpacker_binary .= " -x$ENV{'TMPDIR'}/";
} elsif ($mimeunpacker_binary =~ /ripmime/) {
   $mimeunpacker_binary .= " --unique_names --no-ole --paranoid -i - -d $ENV{'TMPDIR'}/";
}

  
#Get current timestamp for logs
my ($sec,$min,$hour,$mday,$mon,$year,$nowtime);
($sec,$min,$hour,$mday,$mon,$year) = localtime(time);
#my $nowtime = sprintf "%02d/%02d/%02d %02d:%02d:%02d", $mday, $mon+1, $year+1900, $hour, $min, $sec;
#my $nowtime = strftime("%a, %d %b %Y %H:%M:%S %Z", localtime(time));
my ($smtp_sender,$remote_smtp_ip,$remote_smtp_auth,$real_uid,$effective_uid);

$real_uid=$<;
$effective_uid=$>;

if ($DEBUG ) {
  open(LOG,">>$logdir/$debuglog");
  select(LOG);$|=1;
  &debug("+++ starting debugging for process $$ by uid=$real_uid");
}

&debug("setting UID to EUID so subprocesses can access files generated by this script");
$< = $>;            # set real to effective uid
$( = $);            # set real to effective gid

&debug("program name is $prog, version $VERSION");
if ($opt_z) {
  &scan_queue;
  exit 0;
}


&scanner_info;

my (%headers , %virtualheader);
my ($CRYPTO_TYPE,$DOMKEYS,$altered_subject, $HEADERS, $env_returnpath, $returnpath, $QS_RELAYCLIENT);
my ($ATTACHMENT, %BOUNDARY,$BOUNDARY_REGEX,$attachment_header,$attachment_value,%attach_hdrs,%content_type);
my ($ct_attachment_filename,$cd_attachment_filename);
my ($env_recips, $recips, $trecips, $recip, $one_recip);
my ($alarm_status,$elapsed_time,$msg_size,$file_desc);
my ($description,$quarantine_description,$illegal_mime);
my $skip_text_msgs=1;
my $plain_text_msg=0;
my $indicates_attachments=0;
my $xstatus=0;
my $attachment_counter=0;

&working_copy;


#We will set our own value here as it allows us to unset
#it later without changing how Qmail actually interprets
#RELAYCLIENT
$QS_RELAYCLIENT=1 if (defined($ENV{'RELAYCLIENT'}));

if ($ENV{'TCPREMOTEIP'}) {
  $remote_smtp_ip=$ENV{'TCPREMOTEIP'};
  if ($ENV{'TCPREMOTEINFO'}) {
    $remote_smtp_auth=" (".$ENV{'TCPREMOTEINFO'}."\@$remote_smtp_ip)";
    $smtp_sender="via SMTP from $remote_smtp_ip using auth $remote_smtp_auth";
  }else{
    $smtp_sender="via SMTP from $remote_smtp_ip";
  }
  $tag_score="RC:1($remote_smtp_ip):" if ($QS_RELAYCLIENT);
  &debug("incoming SMTP connection from $smtp_sender");
  #system("/usr/bin/printenv > /tmp/qmail-scanner.env");
} else {
  $smtp_sender="via local process $$";
  $remote_smtp_ip='127.0.0.1';
  #Set QS_RELAYCLIENT if QS_SPAMASSASSIN isn't set
  $QS_RELAYCLIENT=1;
  $tag_score="RC:1($remote_smtp_ip):"; #Always would be relayed
  &debug("incoming pipe connection from $smtp_sender");
}
$tag_score="RC:0($remote_smtp_ip):" if ($tag_score !~ /^RC:1/);


#Now alarm this area so that hung networks/virus scanners don't cause 
#double-delivery...

eval {
  $SIG{ALRM} = sub { die "Maximum time exceeded. Something cannot handle this message." };
  alarm $MAXTIME;

  &deconstruct_msg; #JLH if (!$quarantine_event);
  
  
  #Now unset env var QMAILQUEUE so any further Email's sent don't
  #go through the Qmail-Scanner again
  &debug("unsetting QMAILQUEUE env var");
  delete $ENV{'QMAILQUEUE'};
  
  #This SMTP session is incomplete until we see dem envelope headers!
  &grab_envelope_hdrs;
  &debug("from=$headers{'from'},subj=$headers{'subject'}, $qsmsgid=$headers{$qsmsgid} $smtp_sender");

  #Add envelope details to headers array so that they can be matched within
  #perlscanner.
  #Note how they're uppercase cf the message headers which are all forced
  #lowercased. This is to ensure no-one can override them...

  $headers{'MAILFROM'}=$returnpath;
  $headers{'RCPTTO'}=$recips;
  $headers{'REMOTEIPADDR'}=$remote_smtp_ip;

  if ( ($BAD_MIME_CHECKS > 1 && $headers{'mime-version'} eq "") || ($headers{'mime-version'} ne "" && $headers{'content-type'} =~ /^text\/plain/i)) {
    #Hmm, doesn't look nice, but it feels better to make this a separate check for some reason
    if ($skip_text_msgs && ($indicates_attachments < 2) && !@uufile_list && !@attachment_list) {
      &debug("This is a PLAIN text message (because it's either not mime, or is text/plain), skip virus scanners - but not antispam scanners");
      $plain_text_msg=1;
    }
  }
  if ($headers{'MAILFROM'} eq "" || $headers{'subject'} =~ /Returned mail:|Mail Transaction Failed/) {
    &debug("This is a bounce message - better assume there's an attachment in it");
    $plain_text_msg=0;
  }
    

  &init_scanners;
  
  if ($quarantine_event || $quarantine_spam) {
    &debug("unsetting TCPREMOTEIP env var");
    delete $ENV{'TCPREMOTEIP'};
    #Reset locale back to original
    $ENV{'LC_ALL'}=$orig_locale;

    #is this a greylist event?
    if ($quarantine_event=~/gr[ae]ylist/i ) {
      #This text will only be seen by those using the "custom-error"
      #patch. Others will just get a general "qq" temp failure msg.
      &log_event;
      print STDERR "Z$quarantine_event";
      &cleanup;
      exit 82;
    }else{    
      &email_quarantine_report;
    }
  } else {
    &qmail_parent_check;
    &qmail_requeue($env_returnpath,$env_recips,"$scandir/$wmaildir/new/$file_id"); 
  }
  alarm 0;
};

$alarm_status=$@;
if ($alarm_status and $alarm_status ne "" ) { 
  if ($alarm_status eq "Maximum time exceeded. Something cannot handle this message.") {
    &error_condition("ALARM: taking longer than $MAXTIME secs. Requeuing...");
  } else {
    &error_condition("Requeuing: $alarm_status");
  }
}

#Msg has been delivered now, so don't want hangs in this part
#to affect delivery

&log_event;

&cleanup;

($sec,$min,$hour,$mday,$mon,$year) = localtime(time);
#$nowtime = sprintf "%02d/%02d/%02d %02d:%02d:%02d", $mday, $mon+1, $year+1900, $hour, $min, $sec;

&debug("all finished. Total of ",tv_interval ($start_time, [gettimeofday])," secs");

# This is commented out as I'm concerned for people running Q-S behind edge gateways.
#Those boxes would then generate a bounce (as they are not the actual spamming SMTP client)
#if ($destring =~ /SPAM/) {
#  &debug("exit with permanent error as this is high-scored SPAM");
#  exit 111;
#}

exit 0;

############################################################################
# Error handling
############################################################################

#Generate uniq identifiers
sub uniq_id {
  return "$hostname" . time .  __LINE__ . $$;
}


sub log_event {
  if ($log_details) {
    $tag_score .= "$CRYPTO_TYPE:" if ($log_crypto && $CRYPTO_TYPE ne "");
    $tag_score .= "$DOMKEYS:" if ($log_crypto && $DOMKEYS ne "");
    #$virtualheader{'CRYPTODETAILS'}="$CRYPTO_TYPE:$DOMKEYS";
    $tag_score=":$tag_score" if ($tag_score ne "");
    if ($trecips =~ /\0T/) {
      for $recip (split(/\0T/,$trecips)) {
	&log_msg("qmail-scanner",($quarantine_event ne "0" ? "$quarantine_event$tag_score" : "Clear$tag_score"),$elapsed_time,$msg_size,$returnpath,$recip,$headers{'subject'},$headers{$qsmsgid},$file_desc) if ($recip ne "");
      }
    } else {
      #Only one recip
      &log_msg("qmail-scanner",($quarantine_event ne "0" ? "$quarantine_event$tag_score" : "Clear$tag_score"),$elapsed_time,$msg_size,$returnpath,$recips,$headers{'subject'},$headers{$qsmsgid},$file_desc);
    }
  }
}

# Fail with the given message and a temporary failure code.
sub error_condition {
  my ($string,$errcode)=@_;
  $errcode=111 if (!$errcode);
  eval {
    syslog('mail|err',"$V_HEADER-$VERSION:[$file_id] $string");
  };
  if ($@) {
    setlogsock('inet');
    syslog('mail|err',"$V_HEADER-$VERSION:[$file_id] $string");
  }
  if ($log_details ne "syslog") {
    warn "$V_HEADER-$VERSION:[$file_id] $string\n";
  }
  #$nowtime = sprintf "%02d/%02d/%02d %02d:%02d:%02d", $mday, $mon+1, $year+1900, $hour, $min, $sec;
  &debug("error_condition: $V_HEADER-$VERSION: $string");
  close(LOG);
  &cleanup;
  exit $errcode;
}

sub debug {
  my $dnowtime = strftime("%a, %d %b %Y %H:%M:%S %Z", localtime(time));
  print LOG "$dnowtime:$$: ",@_,"\n" if ($DEBUG);
}

sub working_copy {
  my ($hdr,$last_hdr,$value,$num_of_headers,$last_header,$last_value,$attachment_filename);
  select(STDIN); $|=1;
  
  &debug("w_c: mkdir $ENV{'TMPDIR'}");
  mkdir("$ENV{'TMPDIR'}",0750)||&error_condition("$ENV{'TMPDIR'} exists - try again later...");
  chdir("$ENV{'TMPDIR'}")||&error_condition("cannot chdir to $ENV{'TMPDIR'}/");
  if (-f "$scandir/$wmaildir/tmp/$file_id" || -f "$scandir/$wmaildir/new/$file_id") {
    &error_condition("$file_id exists, try again later");
  }
  &debug("w_c: start dumping incoming msg into $scandir/$wmaildir/tmp/$file_id [",&deltatime,"]");
  open(TMPFILE,">$scandir/$wmaildir/tmp/$file_id")||&error_condition("cannot write to $scandir/$wmaildir/tmp/$file_id - $!");
  
  my $still_headers=1;
  my $begin_content='';
  my $still_attachment='';
  my $first_received=0;
  while (<STDIN>) {
    if ($still_headers) {
      $HEADERS .= $_;
      #Catch any naughty illegal header chars here
      if ($BAD_MIME_CHECKS && !$IGNORE_EOL_CHECK && /\r|\0/) {
	$illegal_mime=1;
	&debug("w_c: found CRL/NULL in header - invalid if this is a MIME message");
      }
      #Put headers into array
      if (/^\s+(.*)$/ && $last_hdr) {
	#Hmmm, a continuation...
	$headers{$last_hdr} .= $1 if (!$illegal_mime);
      } elsif (/^([^\s]+)/) {
	#This means it's not a continuation header
	if (!$quarantine_event && $BAD_MIME_CHECKS && ($headers{'mime-version'} ne "") && !/^([^\s]+):(.*)$/) {
	  #Wow - a header (not header+value) that goes onto another line - not likely!
	  $illegal_mime=1;
	  $destring='problem';
	  $quarantine_description="Disallowed breakage found in header name - not valid email";
	  $quarantine_event="Policy:Bad_MIME_Break";
	  $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in message";
	  &debug("w_c: disallowed breakage found in header name ($_) - not valid email");
	  #next;
	} else {
	  /^([^\s]+):(.*)$/;
	  $hdr=$1;
	  $last_hdr=tolower($hdr);
	  $value=$2;
	  $value =~ s/^\s//;
	  if (!$quarantine_event && $BAD_MIME_CHECKS && $headers{'mime-version'} ne "" && $hdr =~ /^[^X].*\(/i) {
	    #Wow - a comment *inside* a standard header name. Only viruses are known to do that
	    #Should we test for [^0-9a-z\_\-\=\+] instead?
	    $illegal_mime=1;
	    $destring='problem';
	    $quarantine_description='Disallowed MIME comment found in header name - not valid email';
	    $quarantine_event="Policy:Bad_MIME_Comment";
	    $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in message";
	    &debug("w_c: $quarantine_description");
	  }
	  $num_of_headers++;
	}
	#Don't let this array grow without bounds...
	if ($num_of_headers < $MAX_NUM_HDRS) {
	  if ($hdr =~ /^to|cc/i && $headers{tolower($hdr)}) {
	    #Special-case the To: and Cc: headers.
	    #Broken mailers generate messages with multiple 
	    #instances of these, so merge them into one...
	    $headers{tolower($hdr)} .= ",$value";
	  } elsif ($hdr =~ /^(from|x-mail|User-Agent|Organi|Received|Message-ID|Subject)/i && $headers{tolower($hdr)}) {
	    #Make sure any multiples of these headers are remembered, so that 
	    #perlscanner checks can see all instances - just wrap em up
	    #into one long line
	    $headers{tolower($hdr)} .= " $value";
	  } elsif ($hdr =~ /^received$/i && !$first_received) {
	    $first_received=1;
	    $value=~/\[([0-9]+\.[0-9]+\.[0-9]+\.[0-9]+)\]\)$/;
	    if ($1 =~ /^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/ && !$ENV{'TCPREMOTEIP'}) {
	      $ENV{'TCPREMOTEIP'}=$1;
	      &debug("TCPREMOTEIP not set - configuring as $ENV{'TCPREMOTEIP'}");
	    }else{
	      #&debug("no need to reset TCPREMOTEIP from $value");
	    }
	  } elsif (!$quarantine_event && $BAD_MIME_CHECKS > 1 && (($headers{'mime-version'} ne "" && tolower($hdr) eq "mime-version") || ($headers{'content-type'} ne "" && tolower($hdr) eq "content-type") || ($headers{'content-transfer-encoding'} ne "" && tolower($hdr) eq "content-transfer-encoding") || ($headers{'content-disposition'} ne "" && tolower($hdr) eq "content-disposition"))) {
	    #Why would a legit message have important MIME headers defined >1 time? It could imply someone is trying to sneak
	    #something past SMTP scanners...
	    #Too much parsing needs to be done to do this correctly - stuff 'em - break the sucker ;-/
	    &debug("Duplicate MIME headers found [$hdr] - renaming");
	    print TMPFILE "$V_HEADER-$VERSION: renamed duplicate MIME headers\n";
	    $_="$V_HEADER-Renamed-$_";
	  } else {
	    #All other headers: the last occurance wins!
	    $headers{tolower($hdr)}=$value;
	  }
	}
      }
      if (/^(\r|\r\n|\n)$/) {
	#headers have finished
	$still_headers=0;
	#Normalize selected headers
	$headers{'subject'}=&normalize_string("Subject:",$headers{'subject'});
	#Try to workaround those nasty broken viruses that produce Content-Type without MIME-Version
	#to get around virus scanners
	if ($headers{'mime-version'} eq "") {
	  #Make sure it's a MIME-style Content-type, Sun used to use Content-type for other purposes...
	  if ($BAD_MIME_CHECKS && $headers{'content-type'} =~ /\//) {
	    print TMPFILE "$V_HEADER-$VERSION: added fake MIME-Version header\nMIME-Version: 1.0\n";
	    $headers{'mime-version'}="1.0";
	    &debug("w_c: added fake MIME-Version header");
	  }
	} elsif ($BAD_MIME_CHECKS > 1 && $headers{'content-type'} eq "") {
	  #OK, now do the same for Content-Type. RFCs state "if no Content-Type present, then it's text/plain"
	  #However, Outlook chooses to read the entire message and "figures out" it's mixed/multipart, etc. 
	  #This'll break that - as it should.
	  #I wonder if I shouldn't just block these instead, the only ones I've seen are either viruses or spam...
	  print TMPFILE "$V_HEADER-$VERSION: added fake Content-Type header\nContent-Type: text/plain\n";
	  $headers{'content-type'}="text/plain";
	  &debug("w_c: added fake Content-Type header");
	}
	  if ( $headers{'mime-version'} ne "" && $headers{'content-type'} =~ /^(\s+|)([^\/\s\(]+)(\s+|)\/(\s+|)([^\/\s\(\;]+)/ ) {
	    $content_type{$attachment_counter}="$2/$5";
	    &debug("w_c: primary Content-Type of $content_type{$attachment_counter} found");
	    if ($log_crypto) {
	      $DOMKEYS="CR:DomKeys(signed)" if ($headers{'domainkey-signature'} ne "");
	      if ($content_type{$attachment_counter} =~ /multipart\/signed/i) {
		$CRYPTO_TYPE="CR:SMIME(signed)" if ($CRYPTO_TYPE eq "" && $headers{'content-type'} =~ /protocol=\"application\/(x\-|)pkcs/i);
		$CRYPTO_TYPE="CR:PGP(signed)" if ($CRYPTO_TYPE eq "" && $headers{'content-type'} =~ /protocol=\"application\/(x\-|)pgp/i);
		&debug("found MIME-based crypto ($CRYPTO_TYPE)");
	      } elsif ($content_type{$attachment_counter} =~ /multipart\/encrypted/i) {
		$CRYPTO_TYPE="CR:PGP(encrypted)" if ($headers{'content-type'} =~ /protocol=\"application\/(x\-|)pgp/i);
		&debug("found MIME-based crypto ($CRYPTO_TYPE)");
	      }elsif ($content_type{$attachment_counter} =~ /application\/(x\-|)pkcs7/i) {
		$CRYPTO_TYPE="CR:SMIME(encrypted)" if ($headers{'content-type'} =~ /application\/(x\-|)pkcs7/i);
		&debug("found MIME-based crypto ($CRYPTO_TYPE)");
	      }
	    }
	  } elsif ($headers{'mime-version'} ne "") {
	    $destring="problem";
	    $illegal_mime=1;
	    $quarantine_description="Disallowed MIME Content-Type found - not valid email";
	    &debug($quarantine_description);
	    $quarantine_event="Policy:Bad_MIME_Type";
	    $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in message";
	  }
	#}
	#if ( $headers{'content-type'} =~ /boundary(\s*)=(|\s+|\s*\")([^\"\;]+)($|\;|\")/i) {
	if ( $headers{'mime-version'} ne "" && $headers{'content-type'} =~ /boundary(\s*)=(|\s+|\s*\")([^\s\"\;]+)($|\;|\")/i) {
	  $BOUNDARY{$attachment_counter}=$3;
	  #if (!$quarantine_event && $BAD_MIME_CHECKS > 1 && ($BOUNDARY{$attachment_counter} =~ /\"|\;/ || $BOUNDARY{$attachment_counter} eq "")) {
	    #&debug("w_c: RFC2046 says boundaries ($BOUNDARY{$attachment_counter}) can't contain such chars [see bcharsnospace]");
	    #$destring="problem";
	    #$illegal_mime=1;
	    #$quarantine_description="Disallowed MIME boundary found - potential virus";
	    #$quarantine_event="Policy:Bad_MIME_Boundary";
	    #$description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in message";
	  #}
	  if (!$quarantine_event && $headers{'mime-version'} ne "" && $BAD_MIME_CHECKS > 1 && ( length($BOUNDARY{$attachment_counter}) == 0 || length($BOUNDARY{$attachment_counter}) > 250)) {
	    #RFC2046 says boundarys are 1-70 chars - making it 250 is being *real* liberal...
	    $destring="problem";
	    $illegal_mime=1;
	    $quarantine_description="Disallowed MIME boundary length found (".length($BOUNDARY{$attachment_counter}).") - not valid email";
	    &debug($quarantine_description);
	    $quarantine_event="Policy:Bad_MIME_Length";
	    $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in message";
	  }
	  #Strip off stuff after semicolon, and escape any odd chars
	  $BOUNDARY{$attachment_counter} =~ s/(\"|\;).*$//g;
	  #$BOUNDARY{$attachment_counter} =~ s/([^a-z0-9=\_])/\\\1/gi;
	  $BOUNDARY{$attachment_counter} =~ s/(\W)/\\$1/g;
	  $BOUNDARY_REGEX=$BOUNDARY{$attachment_counter};
	  &debug("w_c: found a top-level boundary definition of $BOUNDARY{$attachment_counter}");	  
	}
	if ( $headers{'content-type'} =~ /name(|\s+)=(|\s+|\s*\")([^\s\"].*)/i) {
	  $ATTACHMENT=$3;
	  $attachment_counter++;
	  #Strip off stuff after semicolon
	  $ATTACHMENT =~ s/(\"|\;).*$//g;
	  &debug("w_c: found a top-level file attachment definition of $ATTACHMENT");
	  push(@attachment_list, $ATTACHMENT);
	}
	if ($headers{'message-id'} eq "" && !$headers{$qsmsgid}) {
	  $headers{$qsmsgid}="<".time . __LINE__ . $$ . "\@$hostname>";
	  print TMPFILE "${V_HEADER}-Message-ID: $headers{$qsmsgid}\n";
	} else {
	  if (!$headers{$qsmsgid}) {
	    $headers{$qsmsgid}=$headers{'message-id'};
	  }
	}
      }
    }
    if (/^(\r|\r\n|\n)$/) {
      #&debug("w_c: attachment num=$attachment_counter");
      #&debug("w_c: last attachment header: $attachment_header:$attachment_value");
      $attach_hdrs{tolower($attachment_header)}=$attachment_value;
      if ($still_attachment ne "") {
	$still_attachment='';
	$begin_content=$attach_hdrs{'content-transfer-encoding'};
      } else {
	$begin_content='';
      }
      $attachment_header=$attachment_value='';
      #Let's see what the last MIME attachment contained
      if ($cd_attachment_filename ne "" && $ct_attachment_filename ne "" && $ct_attachment_filename ne $cd_attachment_filename) {
	if (!$quarantine_event && $BAD_MIME_CHECKS > 1) {
	  &debug("w_c: Disallowed MIME filename manipulation - potential virus");
	  $illegal_mime=1;
	  $destring="problem";
	  $quarantine_description='Disallowed MIME filename manipulation - not valid email';
	  &debug($quarantine_description);
	  $quarantine_event="Policy:Bad_MIME_Manipulation";
	  $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in message attachment: \"$ct_attachment_filename\" != \"$cd_attachment_filename\"";
	}
      }
      #$ct_attachment_filename=$cd_attachment_filename='';
      if ($attach_hdrs{'content-type'} =~ /name(|\s+)=(|\s+|\s*\")([^\s\"].*)/i && $ATTACHMENT eq "") {
	$ATTACHMENT=$3;
	#Strip off stuff after semicolon
	$ATTACHMENT =~ s/(\"|\;).*$//g;
	$ATTACHMENT=&normalize_string("Filename:",$ATTACHMENT);
	$ATTACHMENT=tolower($ATTACHMENT);
	if (!grep(/^\Q$ATTACHMENT\E$/,@attachment_list)) {
	  &debug("found C-T attachment filename \"$ATTACHMENT\"");
	  push(@attachment_list, $ATTACHMENT);
	}
	$ct_attachment_filename=$ATTACHMENT;
	$ATTACHMENT='';
	#&debug("w_c: found a Content-Type attachment filename of \"$ct_attachment_filename\"");
      }
      if ($attach_hdrs{'content-disposition'} =~ /name(|\s+)=(|\s+|\s*\")([^\s\"].*)/i && $ATTACHMENT eq "") {
	$ATTACHMENT=$3;
	#Strip off stuff after semicolon
	$ATTACHMENT =~ s/(\"|\;).*$//g;
	$ATTACHMENT=&normalize_string("Filename:",$ATTACHMENT);
	$ATTACHMENT=tolower($ATTACHMENT);
	if (!grep(/^\Q$ATTACHMENT\E$/,@attachment_list)) {
	  push(@attachment_list, $ATTACHMENT);
	  &debug("found C-D attachment filename \"$ATTACHMENT\"");
	}
	$cd_attachment_filename=$ATTACHMENT;
	$ATTACHMENT='';
	#&debug("w_c: found a Content-Disposition attachment filename of \"$cd_attachment_filename\"");
      }
      if ($attach_hdrs{'content-type'} =~ /boundary(|\s+)=(|\s+|\s*\")([^\s\"].*)/i) {
	$BOUNDARY{$attachment_counter}=$3;
	#Strip off delimiters around boundary
	$BOUNDARY{$attachment_counter} =~ s/(\"|\;).*$//g;
	$BOUNDARY{$attachment_counter} =~ s/(\W)/\\$1/g;
	if (!$quarantine_event && $headers{'mime-version'} ne "" && $BAD_MIME_CHECKS > 1 && length($BOUNDARY{$attachment_counter}) > 250) {
	  #RFC2046 says boundarys are 0-70 chars
	  $destring="problem";
	  $illegal_mime=1;
	  $quarantine_description="Disallowed MIME boundary length found (".length($BOUNDARY{$attachment_counter}).") - not valid email";
	  &debug($quarantine_description);
	  $quarantine_event="Policy:Bad_MIME_Boundary";
	  $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in message";
	}
	if ( !$quarantine_event && $headers{'mime-version'} ne "" && $BAD_MIME_CHECKS > 1 && $BOUNDARY{$attachment_counter} =~ /^($BOUNDARY_REGEX)$/i) {
	  &debug("w_c: hmm, a new boundary defintion that has already being set. Sounds like a trojan");
          &debug("w_c: broken attachment MIME details - block it!");
          $illegal_mime=1;
	  $destring="problem";
          $quarantine_description='Disallowed MIME boundary found in attachment - not valid email';
	  &debug($quarantine_description);
	  $quarantine_event="Policy:Bad_MIME_Boundary";
	  $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in message";
        }
	if ($BOUNDARY_REGEX ne "") {
	  $BOUNDARY_REGEX.="|".$BOUNDARY{$attachment_counter};
	} else {
	  $BOUNDARY_REGEX=$BOUNDARY{$attachment_counter};
	}
	#&debug("w_c: BOUNDARY_REGEX=$BOUNDARY_REGEX");
      }
      if ($attach_hdrs{'content-type'} =~ /\//) {
	$attachment_filename='';
	$attachment_filename=$cd_attachment_filename ne "" ? $cd_attachment_filename : $ct_attachment_filename;
	#&debug("w_c: just parsed attachment $attach_hdrs{'content-type'}: filename=$attachment_filename");
	if ( $attach_hdrs{'content-type'} =~ /^(\s+|)([^\/\s\(]+)(\s+|)\/(\s+|)([^\/\s\(\;]+)/ ) {
	  $content_type{$attachment_counter}="$2/$5";
	  &debug("w_c: attachment  $attachment_counter: Content-Type of $content_type{$attachment_counter} found");
	  if ($attachment_filename =~ /\.(scr|pif|vbs|exe)$/i && $content_type{$attachment_counter} !~ /^(message|text|application|binary)/i) {
	    $quarantine_description="Disallowed file ($attachment_filename) assosiated with unrelated MIME type ($content_type{$attachment_counter}) - forged attachments blocked";
	    &debug("w_c: $quarantine_description");
	    $illegal_mime=1;
	    $destring='problem';
	    &debug($quarantine_description);
	    $quarantine_event="Policy:Forged_Attachment";
	    $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in attachment $attachment_filename";
	  }
	}
	$attach_hdrs{'content-type'}=$attach_hdrs{'content-disposition'}='';
	$ct_attachment_filename=$cd_attachment_filename='';
      }
    } else {
      #&debug("line=$_");
    }
    
    if ($still_attachment ne "") {
      #&debug("w_c: check those attachment headers ($_)");
      if (/^([^\s]+):(|\s+)(.*)$/) {
	$last_header=$attachment_header;
	$last_value=$attachment_value;
	$attachment_header=$1;
	$attachment_value=$3;
	$attachment_value =~ s/^\s+//;
	if ($last_header) {
	  #&debug("w_c: $last_header:$last_value");
	  $attach_hdrs{tolower($last_header)}=$last_value;
	}
	#&debug("w_c: beginning of $attachment_header, value=$attachment_value");
      } elsif (/^\s(.+)/) {
	#&debug("w_c: line :$_: reached");
	$attachment_value.=$1;
      } elsif (/^(\r|\r\n|\n|\s+)$/) {
	#Yeah - I should block spaces, but too many valid lists send out such junk...
	$still_attachment='';
      } else {
	#This will catch headers that are *correctly* broken over two lines.
	#No known mailer does that, but virus writers do, so we block it.
	#Note that a lot of mailing-lists (and AV systems...) shove their trailers
	#on the bottom of messages irrespective of whether they are MIME or not - so
	#we must allow such "hacks" to slip through
	if (!$quarantine_event && $BAD_MIME_CHECKS > 1 && ($BOUNDARY_REGEX ne "" && $still_attachment !~ /^\-\-($BOUNDARY_REGEX)\-\-$/) ) {
	  &debug("w_c: broken attachment MIME details (still_attachment=$still_attachment, but BOUNDARY_REGEX=\"$BOUNDARY_REGEX\")- block it!");
	  $illegal_mime=1;
	  $destring="problem";
	  $quarantine_description='Disallowed content found in MIME attachment - not valid email';
	  &debug($quarantine_description);
	  $quarantine_event="Policy:Bad_MIME_Header";
	  $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in message";
	}
      }
    }
    if ($begin_content =~ /base64/i && !/^\s/) {
      #&debug("w_c: begin=\"$begin_content\",line=$_");
      $begin_content='';
      #Only looking for base64 encoded as both QP and binary appear to arrive corrupted under Outlook
      if ($_ =~ /^TV(qq|qQ|r1|pQ|pA|py|rm|rh|oF|oI|rQ|o8|ou|oA)/) {
	&debug("w_c: base64 looks like a Windows executable, filename=$attachment_filename,type=$content_type{$attachment_counter}");
	if (!$quarantine_event && $BAD_MIME_CHECKS > 1 && $content_type{$attachment_counter} !~ /^(binary|application)/i) {
	  #As far as I'm aware, a Windows/DOS executable should always be of type "application/<something>"
	  $illegal_mime=1;
	  $destring="problem";
	  $quarantine_description="Disallowed executable attachment associated with \"$content_type{$attachment_counter}\" MIME type - forged attachment";
	  &debug($quarantine_description);
	  $quarantine_event="Policy:Forged_Attachment";
	  $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in attachment \"$attachment_filename\"";
	  &debug("w_c: $quarantine_description");
	}
      }
      if ($_ =~ /^(UEsDB[AB]|UEswMFBL)/) {
	&debug("w_c: base64 looks like a zip file, filename=$attachment_filename,type=$content_type{$attachment_counter}");
	if (!$quarantine_event && $BAD_MIME_CHECKS > 2 && $attachment_filename !~ /\.zip$/i) {
	  #This is a zip file, and yet the filename doesn't end in .zip - should quarantine it!
	  $illegal_mime=1;
	  $destring="problem";
	  $quarantine_description="Disallowed zip attachment when not associated with a .zip filename - forged attachment";
	  &debug($quarantine_description);
	  $quarantine_event="Policy:Forged_Attachment";
	  $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in attachment \"$attachment_filename\"";
	  &debug("w_c: $quarantine_description");
	}
      }
    }
    if ($BOUNDARY_REGEX ne "" && /^\-\-($BOUNDARY_REGEX)/) {
      $still_attachment=$_;
      chomp($still_attachment);
      if (/^\-\-($BOUNDARY_REGEX)\-\-.$/) {
	&debug("w_c: found end of attachment boundary, BOUNDARY_REGEX was \"$BOUNDARY_REGEX\"...");
	my ($delete_bb)=$1;
	$delete_bb =~ s/(\W)/\\$1/g;
	$BOUNDARY_REGEX =~ s/\Q$delete_bb\E//;
	$BOUNDARY_REGEX =~ s/\|\|//;
	$BOUNDARY_REGEX =~ s/(^\||\|$)//;
	&debug("w_c: now that \"$delete_bb\" has been removed, it's \"$BOUNDARY_REGEX\"...");
      }
      $attachment_counter++;
      #&debug("w_c: found :$BOUNDARY_REGEX: - must be attachment section $attachment_counter");
    }
    if ($CRYPTO_TYPE eq "" && $log_crypto) {
      
      $CRYPTO_TYPE="CR:PGP(old-signed)" if (/^(\-\-\-\-\-BEGIN PGP SIGNATURE\-\-\-\-\-|LS0tLS1CRUdJTiBQR1AgU0lHTkFUVVJFLS0tLS0)/);
      $CRYPTO_TYPE="CR:PGP(old-encrypted)" if (/^(\-\-\-\-\-BEGIN PGP MESSAGE\-\-\-\-\-|LS0tLS1CRUdJTiBQR1AgTUVTU0FHRS0tLS0t)/);
      &debug("found old PGP crypto ($CRYPTO_TYPE)") if ($CRYPTO_TYPE ne "");
    }
    &check_and_grab_attachments;
    print TMPFILE ;
  }
  close(TMPFILE)||&error_condition("cannot close $scandir/$wmaildir/tmp/$file_id - $!");

  #scanning message has finished

  $HEADERS =~ s/\r|\0//g;

  &debug("w_c: rename new msg from $scandir/$wmaildir/tmp/$file_id to $scandir/$wmaildir/new/$file_id");
  &debug("w_c: total time between DATA command and \".\" was ",&deltatime," secs");
  &debug("w_c: (this is basically the time it took the client to send the message over the network");
  &debug("w_c: resetting timer so as to measure actual Qmail-Scanner processing time");
  $start_time=[gettimeofday];
  #Not atomic but who cares about the overhead - this is the only app using this area...
  link("$scandir/$wmaildir/tmp/$file_id","$scandir/$wmaildir/new/$file_id")||&error_condition("cannot link $scandir/$wmaildir/tmp/$file_id into $scandir/$wmaildir/new/$file_id - $!");
  unlink("$scandir/$wmaildir/tmp/$file_id")||&error_condition("cannot delete $scandir/$wmaildir/tmp/$file_id - $!");
  my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$atime,$mtime,$ctime,$blksize,$blocks);
  ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$msg_size,$atime,$mtime,$ctime,$blksize,$blocks) = stat("$scandir/$wmaildir/new/$file_id");
  if (!$headers{'date'}) {
    my (@day, @mon);
    $day[0]='Sun';$day[1]='Mon';$day[2]='Tue';$day[3]='Wed';$day[4]='Thu';$day[5]='Fri';$day[6]='Sat';
    $mon[0]='Jan';$mon[1]='Feb';$mon[2]='Mar';$mon[3]='Apr';$mon[4]='May';$mon[5]='Jun';$mon[6]='Jul';$mon[7]='Aug';$mon[8]='Sep';$mon[9]='Oct';$mon[10]='Nov';$mon[11]='Dec';
    my ($tm_sec,$tm_min,$tm_hour,$tm_mday,$tm_mon,$tm_year,$tm_wday,$tm_yday,$tm_isdst);
    ($tm_sec,$tm_min,$tm_hour,$tm_mday,$tm_mon,$tm_year,$tm_wday,$tm_yday,$tm_isdst)=localtime;
    $tm_year += 1900;
    $headers{'date'}=$day[$tm_wday].", $tm_mday ".$mon[$tm_mon]." $tm_year $tm_hour:$tm_min:$tm_sec";
  }
}

sub grab_envelope_hdrs {
  select(STDOUT); $|=1;
  
  open(SOUT,"<&1")||&error_condition("cannot dup fd 0 - $!");
  while (<SOUT>) {
    ($env_returnpath,$env_recips) = split(/\0/,$_,2);
    if ( ($returnpath=$env_returnpath) =~ s/^F(.*)$// ) {
      $returnpath=$1;
      ($recips=$env_recips) =~ s/^T//;
      $recips =~ /^(.*)\0+$/;
      $recips=$1;
      $recips =~ s/\0+$//g;
      #Keep a note of the NULL-separated addresses
      $trecips=$recips;
      $one_recip=$trecips if ($trecips !~ /\0T/);
      $recips =~ s/\0T/\,/g;
    }
    #only meant to be one line!
    last;
  }
  close(SOUT)||&error_condition("cannot close fd 1 - $!");
  if ( ($env_returnpath eq "" && $env_recips eq "") || ($returnpath eq "" && $recips eq "") ) {
    #At the very least this is supposed to be $env_returnpath='F' - so
    #qmail-smtpd must be officially dropping the incoming message for
    #some (valid) reason (including the other end dropping the connection).
    &debug("g_e_h: no sender and no recips. Probably due to SMTP client dropping connection. Nothing we can do - cleanup and exit. This is not necessarily an error!");
    &cleanup;
    exit;
  }
  &debug("g_e_h: return-path is \"$returnpath\", recips is \"$recips\"");
}


sub deconstruct_msg {
  my ($start_decon_time) = [gettimeofday];
  my $save_filename ='';
  my ($new_filename,$MAYBETNEF,$tnef_status);

  my($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat("$scandir/$wmaildir/new/$file_id");

  #override absurd values
  $MAX_SCAN_SIZE=10000000 if ($MAX_SCAN_SIZE < 10000000);
  if ($size > $MAX_SCAN_SIZE)  {
    &debug("d_m: msg is $size bytes - too large to scan");
    $SKIP_SCANNING=1;
  }
  &debug("d_m: starting $mimeunpacker_binary <$scandir/$wmaildir/new/$file_id [",&deltatime,"]");
  open(MIME,"$mimeunpacker_binary  <$scandir/$wmaildir/new/$file_id 2>&1|")||&error_condition("cannot call $mimeunpacker_binary - $!");
  while (<MIME>) {
    next if (/exists/);
    &error_condition("d_m: output spotted from $mimeunpacker_binary ($_) - that shouldn't happen!");
  }
  close(MIME)||&error_condition("cannot close $mimeunpacker_binary - $!");
  my $unpacker='';
  
  opendir(DIR,"$ENV{'TMPDIR'}/")||&error_condition("cannot open dir $ENV{'TMPDIR'}/ - $!"); 
  my @all_unpacked_files = grep(!/^\.+$/, readdir(DIR));
  closedir(DIR);
  &debug("d_m: finished $mimeunpacker_binary [",&deltatime,"]");
  #If you have the tnef app, you'll be able to scan broken M$ attachments
  
  if ( $tnef_binary ) {
    &debug("d_m: Checking all attachments to see if they're MS-TNEF");
    foreach $save_filename (@all_unpacked_files) {
      #Clean up $save_filename so as to keep taint happy
      $save_filename =~ /^(.*)$/; $save_filename=$1;
      ($new_filename=$save_filename) =~ s/([^a-z0-9\.\-\_\+\=\~]+)//gi;
      if ($save_filename ne $new_filename) {
	$new_filename =~ /(\.[^\.]+)$/;
	$new_filename=&uniq_id."$new_filename";
	rename($save_filename,$new_filename);
	&debug("d_m: ren $save_filename to $new_filename");
	$save_filename=$new_filename;
      }
      #Who cares if it is or isn't tnef, just scan it!
      if ($tnef_binary) {
	$MAYBETNEF=`$tnef_binary --number-backups -d $ENV{'TMPDIR'}/ -f $ENV{'TMPDIR'}/$save_filename 2>&1`;	
	$tnef_status=$?;
	&debug("d_m: is $ENV{'TMPDIR'}/$save_filename is a TNEF file?: $tnef_status [",&deltatime,"]");
      }
    }
  }

  &debug("d_m: Check for zip files...");
  #Re-initialize directory listing
  opendir(DIR,"$ENV{'TMPDIR'}/")||&error_condition("cannot open dir $ENV{'TMPDIR'}/ - $!"); 
  @all_unpacked_files = grep(!/^\.+$/, readdir(DIR));
  closedir(DIR);
  foreach $save_filename (@all_unpacked_files) {
    #Clean up $save_filename so as to keep taint happy
    $save_filename =~ /^(.*)$/; $save_filename=$1;
    ($new_filename=$save_filename) =~ s/([^a-z0-9\.\-\_\+\=\~]+)//gi;
    if ($save_filename ne $new_filename) {
      $new_filename =~ /(\.[^\.]+)$/;
      $new_filename=&uniq_id."$new_filename";
      rename($save_filename,$new_filename);
      &debug("d_m: ren $save_filename to $new_filename");
      $save_filename=$new_filename;
    }
    if ( $save_filename =~ /\.(zip|exe)$/i) {
      &unzip_file($save_filename);
    }
  }

  if (!$redundant_scanning) {
    if (-f "$ENV{'TMPDIR'}/$save_filename") {
      system $rm_binary,"-f","$ENV{'TMPDIR'}/$save_filename";
    }
  }
  
  my($decon_time)=tv_interval ($start_decon_time, [gettimeofday]);
  &debug("d_m: unpacking message took $decon_time seconds");
}

sub init_scanners {
  my($start_init_scanners_time)=[gettimeofday];
  &debug("ini_sc: start scanning");
  chdir("$ENV{'TMPDIR'}/");
  
  #Delete original zip'ped attachment as there's no point 
  #in the other scanners double-scanning it - unless $redundant scanning
  #is set....
  if ($redundant_scanning) {
    link("$scandir/$wmaildir/new/$file_id","$ENV{'TMPDIR'}/orig-$file_id");
  }
  &debug("ini_sc: recursively scan the directory $ENV{'TMPDIR'}/");

  #Run AV scanners - even if the message is already going to be quarantined
  #due to some Policy: this way you get the definitive answer as to what is
  #a virus... The exception to this is if it looks like a DoS attack - then
  #don't run the AVs over it - as they may be the ones affected by the DoS...

  &scanloop("virus") if (!$quarantine_DOS && !$SKIP_SCANNING);

  #Only run perlscanner if no reason to quarantine found so far
  &perlscan_scanner if (!$quarantine_event);

  #Finishing with running anti-spam checks (no point in 
  #running antispam over quarantined emails!)
  &scanloop("spam") if (!$quarantine_event && !$SKIP_SCANNING);
  
  chdir("$scandir");

  my($decon_time)=tv_interval ($start_init_scanners_time, [gettimeofday]);
  &debug("ini_sc: scanning message took $decon_time seconds");
}


sub perlscan_scanner {
  #This is most efficient if called from within deconstruct_msg

  my($start_perlscan_time)=[gettimeofday];
  my (%array,$var,$lfile,$filename,$section,$apptype,$save_filename);
  my ($type,$desc,$file,$filepattern,$filepath);
  my ($normalized_hdr,$ps_skipfile,$extension);
  my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks,$fsize);
  my ($attachment_list,$perlscan_time);
  &debug("p_s: starting scan of directory \"$ENV{'TMPDIR'}\"...");
  
  use DB_File;
  
  
  tie (%array, 'DB_File', "$db_filename.db", O_RDONLY, 0600) || &error_condition("cannot open $db_filename.db - $!");

  if (!$quarantine_event && $illegal_mime && $headers{'mime-version'} && $BAD_MIME_CHECKS) {
    $destring="problem";
    $quarantine_description="Disallowed  characters found in MIME headers" if (!$quarantine_description);
    $quarantine_event="Policy:Bad_MIME";
    $description .= "\n---perlscanner results ---\n$destring '$quarantine_description'\n found in message";
  }
  #check out headers against DB...
  
  foreach $var (sort keys(%array)) {
    ($type,$desc)=split(/\t/,$array{$var},2);
    &debug("p_s:  '$var' = '$type' = '$desc'");
    if ($type !~ /^SIZE=(\-|)[0-9]+$/) {
      &debug("p_s:  type is a header!");
      $type =~ s/^Policy\-//gi;
      $var=~s/^[0-9]+://;
      if (!grep(/^$type$/,@virtualheaders_array)) {
	#only force lowercase if they are "real" headers
	$type=tolower($type);
      }else{	
	$headers{$type}=$desc if ($headers{$type} eq "");
      }
      $virtualheader{$type}=$var;
      &debug("p_s: checking for objects containing $type: $var");
      $normalized_hdr=&normalize_string("$type:",$headers{$type});

      #Check headers against the "virtualheaders" from quarantine-events.txt
      if ($headers{$type} ne "" && ($headers{$type} =~ /^$virtualheader{$type}$/ || $normalized_hdr =~ /^$virtualheader{$type}$/i)) {
	$quarantine_description="$desc";
	($quarantine_event=$quarantine_description) =~ s/\s/_/g;
	$quarantine_event="Perlscan:".substr($quarantine_event,0,$QE_LEN);
	$quarantine_event=~s/_$//g;
	$description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in file $ENV{'TMPDIR'}/$file";
	&debug("p_s: something to block! ($quarantine_description)");
	last;
      }
    } else {
      &debug("p_s: type is a size!");
    }
  }

  #opendir(DIR,"$ENV{'TMPDIR'}/")||&error_condition("cannot open dir $ENV{'TMPDIR'}/ - $!"); 
  #@allfiles = grep(!/^\.+$/, readdir(DIR));
  #closedir(DIR);
  open(DIR,"$find_binary $ENV{'TMPDIR'}/ -type f |")||&error_condition("cannot open dir $ENV{'TMPDIR'}/ - $!");
  #append any ORIGINAL uuencoded filenames to this directory array
  #so that perlscanner can match on uuencoded filenames
  my @allfiles=<DIR>;
  close(DIR);

  #merge all crypto details
  my ($CRYPTO_DETAILS)="$CRYPTO_TYPE:$DOMKEYS";


  #Block if "Sensitivity:" header set and yet no sign of encryption used
  if ($headers{'sensitivity'} =~ /private|confidential/i) {
    if ($virtualheader{'ISSENSITIVEANDNOCRYPTO'} ne "" && $CRYPTO_TYPE !~ /encrypted/) {
      $quarantine_description=$headers{'ISSENSITIVEANDNOCRYPTO'};
      ($quarantine_event=$quarantine_description) =~ s/\s/_/g;
      $quarantine_event="Policy:Must_Encrypt";
      $quarantine_event=~s/_$//g;
      $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in message";
      &debug("p_s: something to block! ($quarantine_description)");
    }
    $CRYPTO_TYPE=~s/\)$/,private\)/;
  }

  if ($virtualheader{'CRYPTODETAILS'} ne "" && !$quarantine_event && $CRYPTO_DETAILS ne "") {
    &debug("check crypto characteristics of this message against $virtualheader{'CRYPTODETAILS'}");
    if ($CRYPTO_DETAILS =~ /$virtualheader{'CRYPTODETAILS'}/) {
      $destring='problem';
      $quarantine_description=$headers{'CRYPTODETAILS'};
      $quarantine_event="Policy:No_Crypto";
      $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in file $ENV{'TMPDIR'}/$file";
      return;
    }
  }
  if ($#allfiles > $MAX_NUM_UNPACKED_FILES) {
    &debug("w_c: more than $MAX_NUM_UNPACKED_FILES files found - quarantine");
    $illegal_mime=1;
    $destring='problem';
    $quarantine_description="Too many file components found (".$#allfiles.") - potential DoS";
    $quarantine_event="Policy:Many_Files";
    $quarantine_DOS=$quarantine_event;
    $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in file $ENV{'TMPDIR'}/$file";
    $file_desc .= "too_many:$msg_size\t" if ($file_desc !~ /\Q$file\E:$size\t/);
    return;
  }
  foreach $filepath (@allfiles,@uufile_list,@zipfile_list,@attachment_list) {
    chomp($filepath);
    ($file=$filepath)=~s/^.*\///g;
    #skip files that reformime/ripmime generates.
    #This will potentially allow baddies to smuggle files through
    #by using filenames like this... Nothing can be done about that:-(
    #Reformime generates filenames of the form:
    # 967067231.24320-X.host.name (where X is a number)
    #Ripmime generates filenames of the form:
    # textfileX (where X is a number)
    if ($file =~ /^[0-9]+\.[0-9]+\-[0-9]+\.$hostname|^(orig\-|)$file_id|^textfile[0-9]+/) {
      &debug("p_s: skipping auto-generated file $file");
      $ps_skipfile=1;
    } else {
      &debug("p_s: checking $file against perlscanner database...");
      $ps_skipfile=0;
    }

    if (!$ps_skipfile && $virtualheader{'FILELENGTHTOOLONG'} ne "" && !$quarantine_event && length($file) > 256 && $BAD_MIME_CHECKS > 1 ) {
      #&debug("w_c: majorly long attachment filename found - block it");
      $quarantine_description=$headers{'FILELENGTHTOOLONG'};
      $illegal_mime=1;
      $destring='problem';
      $quarantine_event="Policy:Attach_Length";
      $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in file $ENV{'TMPDIR'}/$file";
      $file_desc .= "$file:$msg_size\t" if ($file_desc !~ /\Q$file\E:$size\t/);
      return;
    }

    #Do the patently obvious filename security checks here
    if ( !$ps_skipfile && $BAD_MIME_CHECKS > 1) {
      #Not as thorough as I'd like - but I got too many false positives doing it more generically... :-(
      #The VALID_WINDOWS_EXTENSIONS is based on double-barrel virii caught in a years worth of Qmail-Scanner
      #logs (gotta love those logs!). Notice that I expressly allow "file.exe.exe" through - as the double-extension
      #doesn't hide anything [just implies a user made a mistake]
      if ($virtualheader{'FILEDOUBLEBARRELED'} ne "" && !$quarantine_event && ($file =~ /(^.*)\.($VALID_WINDOWS_EXTENSIONS)\s*\.($SNEAKY_WINDOWS_EXTENSIONS)$/i) && $file !~ /(\.[a-z0-9]{3})\1|\.pp.\.pp.$/i) {
	$quarantine_description=$headers{'FILEDOUBLEBARRELED'};
	$illegal_mime=1;
	$destring='problem';
	$quarantine_event="Policy:Win_Ext";
	$description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in file $ENV{'TMPDIR'}/$file";
	$file_desc .= "$file:$msg_size\t" if ($file_desc !~ /\Q$file\E:$size\t/);
	return;
      }
      if ($virtualheader{'FILECLSID'} ne "" && !$quarantine_event && $file =~ /\{[0-9a-f]{8}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{4}\-[0-9a-f]{12}\}$/i) {
	$quarantine_description=$headers{'FILECLSID'};
	$destring='problem';
	$quarantine_event="Policy:Win_CLSID";
	$description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in file $ENV{'TMPDIR'}/$file";
	$file_desc .= "$file:$msg_size\t" if ($file_desc !~ /\Q$file\E:$size\t/);
	return;
      }
    }
    if ($file =~ /(^.*)(\.[^\.]+)\.?$/) {
      $extension=tolower($2);
    } else {
      $extension="";
    }
    $lfile = tolower($file);
    &debug("p_s: file $file is lowercased to $lfile and has extension $extension") if (!$ps_skipfile);
    #Stat'ing attachment names from @attachment_list will fail on filenames that reformime rewrites
    #that's OK, as they'll still be picked up via their new filename
    ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks) = stat("$filepath");
    #As you stat virtual files as well as real ones, you can't do this check against virtual files...
    if ($effective_uid ne "" && $uid ne "" && $uid != $effective_uid) {
      $DEBUG=101;
      &error_condition("owner of unpacked file \"$filepath\" (uid=$uid) doesn't match UID of Qmail-Scanner (uid=$effective_uid) - can't expect this to work. Fix whatever is creating files with uid=$uid");
    }
    if ($ino && $file_desc !~ /\Q$file\E:$size\t/) {
      #Sanity check so that the virtual attachments don't get double-counted
      $file_desc .= "$file:$size\t";
    }
    &debug("p_s: compare $lfile (size $size) against perlscanner database") if (!$ps_skipfile);
    if ( $array{$extension} && !$ps_skipfile ) {
      $destring="Disallowed attachment type";
      ($fsize,$quarantine_description) = split(/\t/,$array{$extension},2);
      $attachment_list.="$file:$size,";
    }else{
      foreach $filepattern (keys %array) {
	#&debug("p_s: does \"$filepattern\" match against $lfile?");
	if ( $lfile =~ /^${filepattern}$/i) {
	  #$destring="Disallowed attachment type";
	  ($fsize,$quarantine_description) = split(/\t/,$array{$filepattern},2);
	  $attachment_list.="$file:$size,";
	}
      }
    }
    $fsize=~s/^SIZE=//;
    if (!$ps_skipfile && $quarantine_description && !$quarantine_event && ($size eq $fsize || $fsize =~ /^-1$/i) ) {
      &debug("p_s: Quarantine $file! ($quarantine_description)");
      ($quarantine_event=$quarantine_description) =~ s/\s/_/g;
      if ($quarantine_event=~/gr[ea]ylist/i) {
	$quarantine_event="Perlscan:Greylisted";
      }else{
	$quarantine_event="Perlscan:".substr($quarantine_event,0,$QE_LEN);
      }
      $quarantine_event=~s/_$//g;
      $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in file $ENV{'TMPDIR'}/$file";
      $section=$apptype=$save_filename=$filename="";
      #	return;
    }
  }
  untie %array;


  if ($CRYPTO_TYPE=~/CR:ZIP/ && $virtualheader{'ZIPPASSWORDPROTECTED'} ne "" && !$quarantine_event) {
    $quarantine_description=$headers{'ZIPPASSWORDPROTECTED'};
    &debug("u_f: $quarantine_description");
    $destring='problem';
    $quarantine_event="Policy:Encrypted_ZIP";
    $description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in zip file";
    $file_desc .= "encrypted_zip:$msg_size\t";
    return;
  }

  chdir("$scandir/");
  my($stop_perlscan_time)=[gettimeofday];
  $perlscan_time = tv_interval ($start_perlscan_time, $stop_perlscan_time);
  &debug("p_s:  finished scan of dir \"$ENV{'TMPDIR'}\" in $perlscan_time secs");
}


sub scanloop {
  my($scanType)=@_;
  &debug("scanloop($scanType): starting scan of directory \"$ENV{'TMPDIR'}\"...");
  
  my ($scanner);
  #Remember any policy blocks that have already occurred, but reset
  #$quarantine_event so that if a virus is found, that "wins"
  $quarantine_event_tmp=$quarantine_event;
  $quarantine_event='0';
  foreach $scanner (@scanner_array) {
    #Any scanner errors caused by broken zip files/etc will be ignored
    # - not sure how that should be handled...

    #Just run virus scanners over mail that isn't plain text
    if ($plain_text_msg) {
      #If it's plain text - just run anti-spam checks
      if ($scanType eq "spam" && $scanner =~ /(^spam)/i) {
	&debug("scanloop: scanner=$scanner,plain_text_msg=$plain_text_msg");
	&{$scanner};
      }
    }else {
      if ($scanType ne "spam") {
	if ($scanner !~ /(^spam)/i) {
	  &debug("scanloop: scanner=$scanner,plain_text_msg=$plain_text_msg");
	  &{$scanner};
	}
      }else{
	if ($scanner =~ /(^spam)/i) {
	  &debug("scanloop: scanner=$scanner,plain_text_msg=$plain_text_msg");
	  &{$scanner};
	}
      }	
    }
    if ($quarantine_event) {
      #If one scanner finds a virus - why run the rest over it?
      #Make sure this is set correctly
      $destring="virus" if (!$destring);
      last;
    }
  }
  &debug("scanloop: finished scan of \"$ENV{'TMPDIR'}\"...");
  #Reset back the state if these scanners haven't overridden it
  $quarantine_event=$quarantine_event_tmp if (!$quarantine_event);
}

sub qmail_requeue {
  my($sender,$env_recips,$msg)=@_;
  my ($temp,$findate);

  &debug("q_r: fork off child into $qmailqueue...");
  
  #($recips=$env_recips) =~ s/^T//;
  #$recips =~ s/\0T/\,/g;
  #$recips =~ /^(.*)\0+$/;
  #$recips = $1;
  #$recips =~ s/\0+$//g;
  
  # Create a pipe through which to send the envelope addresses.
  pipe (EOUT, EIN) or &error_condition("Unable to create a pipe. - $!");
  select(EOUT);$|=1;
  select(EIN);$|=1;
  # Fork qmail-queue.  The qmail-queue child will then open fd 0 as
  # $message and fd 1 as the reading end of the envelope pipe and exec
  # qmail-queue.  The parent will read in the addresses and pass them 
  # through the pipe and then check the exit status.

  $elapsed_time = tv_interval ($start_time, [gettimeofday]);
  local $SIG{PIPE} = 'IGNORE';
  my $pid = fork;

  if (not defined $pid) {
    &error_condition ("Unable to fork. (#4.3.0) - $!");
  } elsif ($pid == 0) {
    # In child.  Mutilate our file handles.
    close EIN; 
    
    open(STDIN,"<$msg")|| &error_condition ("Unable to reopen fd 0. (#4.3.0) - $!");

    open (STDOUT, "<&EOUT") ||  &error_condition ("Unable to reopen fd 1. (#4.3.0) - $!");
    select(STDIN);$|=1;
    &debug("q_r: xstatus=$xstatus");
    open (QMQ, "|$qmailqueue")|| &error_condition ("Unable to open pipe to $qmailqueue [$xstatus] (#4.3.0) - $!");
    ($sec,$min,$hour,$mday,$mon,$year) = gmtime(time);
    $elapsed_time = tv_interval ($start_time, [gettimeofday]);
    $findate = POSIX::strftime( "%d %b ",$sec,$min,$hour,$mday,$mon,$year);
    $findate .= sprintf "%02d %02d:%02d:%02d -0000", $year+1900, $hour, $min, $sec;
    print QMQ "Received: from $remote_smtp_ip$remote_smtp_auth by $hostname (envelope-from <$returnpath>, uid $real_uid) with qmail-scanner-$VERSION \n";
    print QMQ " ($SCANINFO \n Clear:$tag_score. \n";
    print QMQ " Processed in $elapsed_time secs); $findate\n";
    if ($sa_comment ne "") {
      print QMQ "X-Spam-Status: $sa_comment\n"; 
      print QMQ "X-Spam-Level: $sa_level\n";
    }
    #Only add these headers for Internet-incoming
    if ( $descriptive_hdrs && !$QS_RELAYCLIENT) {
      print QMQ "${V_HEADER}-Mail-From: $returnpath via $hostname\n";
      print QMQ "${V_HEADER}-Rcpt-To: $recips\n" if ($descriptive_hdrs eq "2");
      print QMQ "$V_HEADER: $VERSION (Clear:$tag_score. Processed in $elapsed_time secs)\n";
    }
    my $still_headers=1;
    my $seen_env=0;
    while (<STDIN>) {
      if ($still_headers && $spamc_options =~ / \-c /) {
	#break any X-Spam-Status/Level IFF we've set a SA value ourselves. Easier than removing - and it leaves
	#them around for diagnosis...
	if ($sa_comment ne "" && /^(X-Spam-Status|X-Spam-Flag|X-Spam-Level):/i) {
	  s/^(X-Spam-Status|X-Spam-Flag|X-Spam-Level):/${V_HEADER}-MOVED-$1:/i;
	}
	if ($sa_comment =~ /^yes/i && $spamc_subject ne "" && !/^Subject: \Q$spamc_subject\E/i && /^(Subject):(\s?)([^\n]+)\n/i ) {
	  $altered_subject="$1: $spamc_subject $3";
	  if ($altered_subject !~ /^: \Q$spamc_subject\E/) {
	    &debug("altering subject line to $altered_subject");
	    print QMQ "$altered_subject\n";
	    next;
	  }
	}
	$still_headers=0 if (/^(\r|\r\n|\n)$/);
        #Insert Subject: line if e-mail dosn't contain one but must be tagged
        print QMQ "Subject: $spamc_subject\n" if ((!$still_headers) && ($sa_comment =~ /^yes/i) && (!$altered_subject) && $spamc_subject ne "" );

      }
      print QMQ;
    }
    close(QMQ); #||&error_condition("Unable to close pipe to $qmailqueue (#4.3.0) - $!");
    $xstatus = ( $? >> 8 );
    if ( $xstatus > 10 && $xstatus < 41 ) {
      &error_condition("mail server permanently rejected message. (#5.3.0) - $!",$xstatus);
    } elsif ($xstatus > 0) {
      &error_condition("Unable to open pipe to $qmailqueue [$xstatus] (#4.3.0) - $!",$xstatus);
    }
    #This child is finished - exit
    exit;
  } else {
    # In parent.
    close EOUT;
      
    # Feed the envelope addresses to qmail-queue.
    print EIN "$sender\0$env_recips";
    close EIN  || &error_condition ("Write error to envelope pipe. (#4.3.0) - $!");
}

  # We should now have queued the message.  Let's find out the exit status
  # of qmail-queue.
  waitpid ($pid, 0);
  $xstatus =($? >> 8);
  if ( $xstatus > 10 && $xstatus < 41 ) {
    &error_condition("mail server permanently rejected message. (#5.3.0) - $!",$xstatus);
  } elsif ($xstatus > 0) {
    &error_condition("Unable to close pipe to $qmailqueue [$xstatus] (#4.3.0) - $!",$xstatus);
  }
}


sub valid_virus_to_report {
  my ($virus_type)=@_;
  my ($virus)='';
  # This subroutine is used to determine if the virus found during the scan
  # is reportable. i.e. do we want to send a message to this user or not as is
  # the case with the KLEZ virus.
  #&debug("v_v_t_r: called with $virus_type");
  foreach $virus (@silent_viruses_array) {
    #&debug("v_v_t_r: does $virus_type contain $virus?");
    if ($virus_type =~ /$virus/i) {
      &debug("v_v_t_r: $virus_type contain $virus - so don't notify the sender");
      return 0;
    }
  }
  return 1;
}

sub automated_msg {
  if ($headers{'x-loop'} || $headers{'auto-submitted'} !~ /^(|no)$/i || $headers{'x-listname'} || $headers{'x-listmember'} || $headers{'mailing-list'} || $headers{'x-mailing-list'} || $headers{'precedence'} =~ /^(bulk|list|junk)$/i || $returnpath =~ /^$|^\#\@\[\]$|anonymous|nobody|daemon|request|bounce|mailer|postm|owner|list|words|majordom|experts|\-(return|error)/i) {
    return 1;
  } else {
    return 0;
  }
}

sub bounce_msg {
  if ($returnpath =~ /^$|^\#\@\[\]$|(daemon|bounce|mailer|postm)/i) {
    return 1;
  } else {
    return 0;
  }
}

sub is_unreplyable_email {
  my ($addr_type)=@_;
  my ($dom,$is_local)='';
  #This subroutine is used to see if the sender of this message
  #was a mailing-list/postmaster/etc, or the recipient is a local user. 
  #If it is we don't want to send a reply.
  #&debug("i_u_e: called with $addr_type");
  
  if ($addr_type eq "recips") {
    foreach $dom (@local_domains_array) {
      #&debug("i_u_e: does $recips contain $dom?");
      if ($recips =~ /$dom$/i) {
	#&debug("i_u_e: yes it does!");
	$is_local++;
      }
    }
  } else {
    $is_local="99";
    if (&automated_msg ) {
      #&debug("i_u_e: $addr_type is a mailing-list");
      return 1;
    }
  }
  #
  #Only reply if it is a local address
  if (!$is_local) {
    #&debug("i_u_e: is_local=$is_local");
    return 1;
  } else {
    #&debug("i_u_e: is_local=$is_local");
    return 0;
  }
}

sub email_quarantine_report {
  my($start_email_time)=[gettimeofday];
  if ($quarantine_spam) {
    #Use a different maildir for SPAM
    $vmaildir=$smaildir;
    $quarantine_event=$quarantine_spam;
  }elsif ($quarantine_event =~ /^(Policy|Perlscan)/) {
    $destring="policy-violation";
    #Use a different maildir for Policy-blocks
    $vmaildir=$pmaildir;
  }
  if ($vmaildir ne "none") {
    &debug("e_v_r: quarantine  msg to $scandir/quarantine/$vmaildir/new/$file_id");
    link("$scandir/$wmaildir/new/$file_id","$scandir/quarantine/$vmaildir/new/$file_id")||&error_condition("cannot link $scandir/$wmaildir/new/$file_id into $scandir/quarantine/$vmaildir/new/ - $!");
  }
  open(QTINE,">>$scandir/quarantine/$vmaildir/new/$file_id");
  print QTINE "\n*** Qmail-Scanner Quarantine Envelope Details Begin ***\n";
  print QTINE "${V_HEADER}-Mail-From: \"$returnpath\" via $hostname\n";
  print QTINE "${V_HEADER}-Rcpt-To: \"$recips\"\n";
  print QTINE "$V_HEADER: $VERSION ($SCANINFO $destring Found. Processed in ",tv_interval($start_time,[gettimeofday])," secs)\n";
  print QTINE "Quarantine-Description: $quarantine_description\n";
  print QTINE "*** Qmail-Scanner Envelope Details End ***\n";
  close QTINE;

  if (!$quarantine_spam) {
    &email_sender("sender") if (&valid_virus_to_report($quarantine_description));
    &email_sender("admin");
    if ($trecips =~ /\0T/) {
      for $recip (split(/\0T/,$trecips)) {
	&email_recips($recip);
      }
    } else {
      &email_recips($recips);
    }
  } else {
    #This is almost 100% certainly SPAM - no point in notifying anyone
  }
  &write_quarantine_report;
  $elapsed_time = tv_interval ($start_time, [gettimeofday]);
  &debug("e_v_r: email_quarantine_report took ".tv_interval ($start_email_time, [gettimeofday])." seconds to execute");
}

sub cleanup {
  closelog();
  chdir("$scandir/");
  if ($archiveit !~ /^(1|yes)$/i) {
    #This will only archive mail where the sender or recipient matches the regex that is $archiveit
    if ($headers{'MAILFROM'} !~ /$archiveit/i && $headers{'RCPTTO'} !~ /$archiveit/i) {
      $archiveit=0;
    }
  }
  if (!$archiveit) {
    &debug("cleanup: $rm_binary -rf $ENV{'TMPDIR'}/ $scandir/$wmaildir/new/$file_id") ;
  } else {
    # check if $archivedir exists
    if (!-d "$scandir/$archivedir") {
      mkdir("$scandir/$archivedir",0750) || &error_condition("cannot create $scandir/$archivedir - $!");
      mkdir("$scandir/$archivedir/new",0750) || &error_condition("cannot create $scandir/$archivedir/new - $!");
      mkdir("$scandir/$archivedir/cur",0750) || &error_condition("cannot create $scandir/$archivedir/cur - $!");
      mkdir("$scandir/$archivedir/tmp",0750) || &error_condition("cannot create $scandir/$archivedir/tmp - $!");
    }
    if ( -f "$scandir/$wmaildir/new/$file_id" ) {
      &debug("cleanup: archiving into $scandir/$archivedir/new/");
      rename("$scandir/$wmaildir/new/$file_id","$scandir/$archivedir/new/$file_id");
      #This will do for now. Not pretty - but very cheap!
      #We need to append this information, otherwise how do you know who this message
      #was from or to?
      #
      open(ARCHIVE,">>$scandir/$archivedir/new/$file_id");
      print ARCHIVE "\n*** Qmail-Scanner Envelope Details Begin ***\n";
      print ARCHIVE "${V_HEADER}-Mail-From: \"$returnpath\" via $hostname\n";
      print ARCHIVE "${V_HEADER}-Rcpt-To: \"$recips\"\n";
      print ARCHIVE "$V_HEADER: $VERSION ($SCANINFO  Clear:$tag_score. Processed in ",tv_interval($start_time,[gettimeofday])," secs)\n";
      print ARCHIVE "*** Qmail-Scanner Envelope Details End ***\n";
      close ARCHIVE;
    } 
  }
  if ($DEBUG < 100 && $file_id ne "") {
    chdir("$scandir/");
    system("$rm_binary -rf $ENV{'TMPDIR'}/ $scandir/$wmaildir/new/$file_id");
  }

}


sub scan_queue {
  my ($scanner,$SCANINFO,$files,$sweep_eng,$sweep_product,$sophie_eng,$dir);
  my $start_scan_time =time;
  my ($inocucmd_eng,$inocucmd_product,$spamassassin_eng);
  
  chdir($scandir);
  &debug("s_q: re-create the quarantine version file");
  foreach $scanner (@scanner_array) {
    $scanner =~ s/_scanner//;
    &debug("s_q: detecting version of $scanner");
    if ($scanner eq "uvscan") {
      open(UV,"$uvscan_binary --version|")||die "failed to call $uvscan_binary --version - $!";
      while (<UV>) {
	chomp;
	if (/^Scan engine (v[0-9\.]+) /) {
	  $SCANINFO .="uvscan: $1/";
	} elsif (/^Virus data file (v[0-9\.]+) /) {
	  $SCANINFO .= "$1. ";
	}
      }
      close(UV);
    } elsif ($scanner eq "csav") {
      open(CS,"$csav_binary -virno|")||die "failed to call $csav_binary -virno - $!";
      while (<CS>) {
	chomp;
	if (/Command Software AntiVirus for Linux (version [0-9\.]+)/) {
	  $SCANINFO .="csav: $1/";
	} elsif (/^CSA[V]: (.*)/) {
	  $SCANINFO .= "$1/";
	}
      }
      close(CS);
    } elsif ($scanner eq "trophie" ) {
      open(IS,"$trophie_binary -v 2>&1|")||die "failed to call $trophie_binary -v - $!";
      while (<IS>) {
	chomp;
	if (/VSAPI version (.*)/) {
	  $SCANINFO .= "trophie: $1/";
	} elsif (/Pattern version ([0-9]+) \(pattern number ([0-9]+)\)/) {
	  $SCANINFO .= "$1/$2. ";
	}
      }
      close(IS);
    } elsif ($scanner eq "iscan") {
      open(IS,"$iscan_binary -v|")||die "failed to call $iscan_binary -v - $!";
      while (<IS>) {
	chomp;
	if (/Virus Scanner (v[0-9\.]+), VSAPI (v[0-9\.\-]+)/) {
	  $SCANINFO .="iscan: $1/$2/";
	} elsif (/Pattern version ([0-9\.]+)/) {
	  $SCANINFO .= "$1/";
	} elsif (/Pattern number ([0-9\.]+)/) {
	  $SCANINFO .= "$1. ";
	}
      }
      close(IS);
    } elsif ($scanner eq "fsecure") {
      open(FS,"$fsecure_binary --version|")||die "failed to call $fsecure_binary --version - $!";
      while (<FS>) {
        chomp;
        if (/^F-Secure.*(Release|version)\s+([0-9\.]+)\s+build\s+([0-9]+)/i) {
          $SCANINFO .="fsecure: $2/$3/";
        } elsif (/sign.def version ([0-9\.]+-[0-9\.]+-[0-9\.]+)/) {
          $SCANINFO .= "$1/";
        } elsif (/fsmacro.def version ([0-9\.]+-[0-9\.]+-[0-9\.]+)/) {
          $SCANINFO .= "$1/";
        } elsif (/sign2.def version ([0-9\.]+-[0-9\.]+-[0-9\.]+)/) {
          $SCANINFO .= "$1. ";
        } elsif (/F-PROT database version (.*)$/) {
         $SCANINFO .= "fprot($1)/";
       } elsif (/AVP FPI Engine database version (.*)$/) {
         $SCANINFO .= "avp($1). ";
       } elsif (/Libra database version ([0-9\.]+-[0-9\.]+-[0-9\.]+)/) {
         $SCANINFO .= "libra database $1/";
       } elsif (/Orion database version ([0-9\.]+-[0-9\.]+-[0-9\.]+)/) {
         $SCANINFO .= "orion database $1/";
       } elsif (/AVP FPI Engine database version ([0-9\.]+-[0-9\.]+-[0-9\.]+)/) {
         $SCANINFO .= "avp fpi database $1. ";
       }
      }
      close(FS);
      $SCANINFO .= ". " if ($SCANINFO !~ /\. $/);
    } elsif ($scanner eq "fprot") { 
      open(FP,"$fprot_binary \?|")||die "failed to call $fprot_binary --version - $!";
      while (<FP>) {
	chomp;
	if (/(F-PROT|Program version:) ([0-9\.]+)/) {
	  $SCANINFO .="f-prot: $2/";
	} elsif (/Engine version: ([0-9\.]+)/) {
	  $SCANINFO .= "$1";
	}
      }
      $SCANINFO .= ". ";
      close(FP);
    } elsif ($scanner eq "hbedv") {
      open(IS,"$hbedv_binary --version 2>&1 |")||die "failed to call  $hbedv_binary --version - $!";
      while (<IS>) {
	chomp;
	if (/engine version:\s+([0-9\.]+)/) {
	  $SCANINFO .= "hbedv: $1";
	} elsif (/vdf version:\s+([0-9\.]+)/) {
	  $SCANINFO .= "/$1. ";
	}
      }
      close(IS);
    } elsif ($scanner eq "avp") {
      open(AVP,"$avp_binary -Y -VL 2>&1 |")||die "failed to call  $avp_binary -Y -VL  - $!";
      while (<AVP>) {
	chomp;
	if (/Version (([0-9\.]+)\s+build ([0-9\.]+)|([0-9\.]+))/) {
	  if ($2) {
	    $SCANINFO .= "avp: $1/$2. ";
	  } else {
	    $SCANINFO .= "avp: $1. ";
	  }
	}
      }
      close(AVP);
    } elsif ($scanner eq "ravlin") {
      open(RAV,"$ravlin_binary --version 2>&1 |")||die "failed to call  $ravlin_binary --version  - $!";
      while (<RAV>) {
	chomp;
	if (/^Version: ([0-9\.]+)\./) {
	  $SCANINFO .= "ravlin: $1. ";
	}
      }
      close(RAV);
    } elsif ($scanner eq "vexira") {
      open(VEX,"$vexira_binary --version 2>&1 |")||die "failed to call  $vexira_binary --version  - $!";
      while (<VEX>) {
	chomp;
	if (/^engine version:\s+([0-9\.]+)/) {
	  $SCANINFO .= "vexira: $1. ";
	}
      }
      close(RAV);
    } elsif ($scanner eq "bitdefender") {
      open(BITDEF,"$bitdefender_binary --info 2>&1 |")||die "failed to call $bitdefender_binary --info - $!";
      while(<BITDEF>) {
	chomp;
	if (/^BDC\/Linux\-Console (.*) \(build ([^\)]+)\)/){
	  $SCANINFO .= "bitdefender: $1/$2";
	}
	if (/^Engine signatures:\s+([0-9]+)/) {
	  $SCANINFO .= "/$1. ";
	}
      }
      close(BITDEF);
    } elsif ($scanner eq "nod32") {
      open(NOD,"$nod32_binary --version 2>&1 |")||die "failed to call $nod32_binary --version - $!";
      while(<NOD>) {
	chomp;
	if (/\s(\d\S+)\s*$/){
	  $SCANINFO .= "nod32: $1 ";
	}
      }
      close(NOD);
    } elsif ($scanner eq "sophie") {
      open(SOP,"$sophie_binary -v 2>&1|")||die "failed to call $sophie_binary -v - $!";
      while (<SOP>) {
	chomp;
	if (/Sophos engine version (.*)$/) {
	  $sweep_eng=$1;
	} elsif (/Sophos IDE version ([0-9\.]+)/) {
	  $sweep_product=$1;
	} elsif (/Sophie version\s+:\s+([0-9\.]+)/) {
	  $sophie_eng=$1;
	}
      }
      $SCANINFO .= "sophie: $sophie_eng/$sweep_eng/$sweep_product. ";
      close(SOP);
    } elsif ($scanner eq "sweep") {
      open(SOP,"$sweep_binary -v|")||die "failed to call $sweep_binary -v - $!";
      while (<SOP>) {
	chomp;
	if (/Engine version\s+:\s+(.*)$/) {
	  $sweep_eng=$1;
	} elsif (/Product version\s+:\s+(.*)$/) {
	  $sweep_product=$1;
	}
      }
      $SCANINFO .= "sweep: $sweep_eng/$sweep_product. "; 
      close(SOP);
    } elsif ($scanner eq "inocucmd") {
      open(IOP,"$inocucmd_binary -HEL|")||die "failed to call $inocucmd_binary -HEL - $!";
      while (<IOP>) {
	chomp;
	if (/Engine version:\s+(.*) ([0-9\/]+)$/) {
	  $inocucmd_eng=$1;
	} elsif (/Data version:\s+(.*) ([0-9\/]+)$/) {
	  $inocucmd_product=$1;
	}
      }
      $SCANINFO .= "inocucmd: $inocucmd_eng/$inocucmd_product. ";
      close(IOP);
    } elsif ($scanner eq "clamscan") {
      open(CLAMS,"$clamscan_binary --stdout -V|")||die "failed to call $clamscan_binary --stdout -V - $!";
      while (<CLAMS>) {
	chomp;
	if (/ersion ([0-9\.\-a-z]+)/i) {
	  $SCANINFO .="clamscan: $1. ";
	}
      }
      close(CLAMS);
    } elsif ($scanner eq "clamdscan") {
      open(CLAMS,"$clamdscan_binary --version 2>&1|")||die "failed to call $clamdscan_binary --version - $!";
      while (<CLAMS>) {
	chomp;
	if (/ersion ([0-9\.\-a-z]+)/i) {
	  $SCANINFO .="clamdscan: $1. ";
	} elsif (/^ClamAV ([^\/]+)\/([^\/]+)\//) {
	  $SCANINFO .="clamdscan: $1/$2. ";
	}
      }
      close(CLAMS);
    } elsif ($scanner eq "spamassassin") {
      #X-Spam-Checker-Version: SpamAssassin 2.01
      open(SPAS,"$spamassassin_binary  -V |")||die "failed to call $spamassassin_binary -V - $!";
      $spamassassin_eng="2.x";
      while (<SPAS>) {
	chomp;
	if (/^SpamAssassin version (.*)$/i) {
	  $spamassassin_eng=$1;
	}
      }
      close(SPAS);
      $SCANINFO .= "spamassassin: $spamassassin_eng. ";
    } else {
      #Catch-all for other ones
      $SCANINFO .= "$scanner: ???. ";
    }
  }
  $SCANINFO =~ s/ \. / /g;
  open(VER,">$versionfile.tmp")||die "cannot write to $versionfile.tmp - $!";
  print VER $SCANINFO;
  close(VER);
  rename("$versionfile.tmp","$versionfile");
  &debug("s_q: cleaning up files older than 2 days via $find_binary $scandir/tmp -mtime +2 -exec $rm_binary -rf {} \;");
  my ($OLDFILES)=`$find_binary $scandir/tmp -mtime +2 -exec $rm_binary -rf {} \\; 2>/dev/null`;
  &debug("s_q: cleaning up quarantined mail older than 14 days via $find_binary $scandir/quarantine -type f -mtime +14 -exec $rm_binary -rf {} \;");
  $OLDFILES=`$find_binary $scandir/quarantine/ -type f -mtime +14 -exec $rm_binary -f {} \\; 2>/dev/null`;
}

sub write_quarantine_report {
  my ($temp,$desc,$report,$subj);
  $subj=$headers{'subject'};
  $subj =~ s/\t/ /g;
  $desc=$quarantine_description;
  $desc =~ s/\n\t/ /g;
  $nowtime = strftime("%a, %d %b %Y %H:%M:%S %Z", localtime(time));
  $report = "$nowtime\t$returnpath\t$recips\t$subj\t$desc\t$SCANINFO\n";
  open(QUARANTINELOG,">>$scandir/$quarantinelog");
  print QUARANTINELOG $report;
  close QUARANTINELOG;
  &debug("w_v_r: writing quarantine log report of: $report");
}

sub scanner_info {
  open(SC,"<$versionfile")||&error_condition("cannot open $versionfile - did you initialise the system by running \"$prog -z\"? - $!");
  $SCANINFO = <SC>;
  $SCANINFO =~ s/\n|\r|\0/ /g;
  close(SC);
}

sub generate_quarantine_db {
  use DB_File;
  use vars qw( %h);
  my ($line,%array,$count,$match,$type,$descr,$entry,$descrip,$size);
  if ($opt_g) {
    print "perlscanner: generate new DB file from $db_filename.txt\n";
    unlink("$db_filename.db.tmp");
    tie (%array, 'DB_File', "$db_filename.db.tmp", O_CREAT|O_RDWR, 0640, $DB_HASH ) || &error_condition("cannot open for write $db_filename.db.tmp - $!");
    
    open(TXT,"<$db_filename.txt")||&error_condition("cannot read $db_filename.txt - $!");

    #Remeber: all filenames are lowercased, but headers aren't...
    while (<TXT>) {
      $line++;
      next if (/^\#|^\s+$/); #ignore lines starting with hashes
      chomp;
      $count++;
      ($match,$type,$descr)=split(/\t+/,$_,3);
      if ( $match eq "" || ($type !~ /^SIZE=(\-|)[0-9]+$/ && $type !~ /^Policy-[0-9a-z\_\-]+:$/i) ) {
	print "ERROR: incorrect format on line \"$line\"\n";
	&error_condition("ERROR: incorrect format on line \"$line\"");
      } else {
	#Strip off any regex endings
	if ($type =~ /^SIZE=(|\-)[0-9]+$/) {
	  #this is a filename/attachment
	  if ( $match =~ /^\.dat$/i ) {
	    
	    print "ERROR: on line \"$line\".\nCannot block all .dat files. Will block too many normal messages.\n";
	    &error_condition("ERROR: on line \"$line\".\nCannot block all .dat files. Will block too many normal messages.");
	    next;
	  }
	  $match = tolower($match);
	} else {
	  #this is for header matches
	  $match =~ s/^\^|\$$//g;
	  #Now make unique
	  $match = "$line:$match";
	  $type =~ s/:$//;
	}
	$array{"$match"}="$type\t$descr";
      }
    }
    close(TXT);
#    $array->sync;
    untie %array ;
    rename("$db_filename.db.tmp","$db_filename.db");
    print "perlscanner: total of $count entries.\n";
  } else {
    print "perlscanner: reading from $db_filename.db\n";
    tie (%array, 'DB_File', "$db_filename.db", O_RDONLY, 0600) || &error_condition("cannot open $db_filename.db - $!");
    foreach $entry (keys %array) {
      $count++;
      ($type,$descrip)=split(/\t/,$array{$entry},2);
      if ( $type =~ /^SIZE=((\-|)[0-9]+)/) {
	if ($type eq "SIZE=-1") {
	  $type="Any";
	} elsif ($type =~ /^SIZE=((\-|)[0-9]+)$/) {
	  $type="$1 bytes";
	}
	print "File: \t$entry\n\t\t\tSize: $type\n\t\t\tDescription: $descrip\n\n";
      }
      if ($type =~ /^Policy-(.*)$/i) {
	$type=$1;
	#Strip off numeric uid...
	$entry =~ s/^[0-9]+://;
	if (grep(/^$type$/,@virtualheaders_array)) {
	  print "Virtual Header: \t$type\n\t\t\tContent: ^$entry\$\n\t\t\tDescription: $descrip\n\n";
	} else {
	  print "Email Header: \t$type\n\t\t\tContent: ^$entry\$\n\t\t\tDescription: $descrip\n\n";
	}
      }
    }
    untie %array;
    print "perlscanner: total of $count entries found.\n";
  }
}




sub show_version {
  my ($scanner);
  &scanner_info;
  print "

$prog

Version: $VERSION

Perl:    Summary of my perl5 (revision 5 version 8 subversion 8) configuration:

Scanners: perlscanner";
  foreach $scanner (@scanner_array) {
    print ", $scanner";
  }

  print "\n\nScanner versioning: $SCANINFO\n";
  print "
Operating System: FreeBSD, 7.1-RELEASE
Hardware:         amd64";
  print "\n\n\n";
}


sub email_sender {
  #Don't e-mail bounced mail messages/etc!
  return if (&is_unreplyable_email('sender'));
  my($addr_type)=@_;
  my ($HDR,$hdr,$tmpsndrs,$tmpsubj,$polstring)='';
  my ($tmpmsgid)= &uniq_id() . "-" . $V_FROM;
  $polstring='policy' if (&notify_addr('nmlvadm'));

  open(SM,"|$qmailinject -h -f ''")||&error_condition("cannot open $qmailinject for sending quarantine report - $!");
  print SM "From: \"$V_FROMNAME\" <$V_FROM>\n";
  if ($addr_type =~ /sender/) {
    $addr_type='psender' if ($NOTIFY_ADDRS =~ /psender/);
    if ($addr_type eq "sender") {
      if (!&is_unreplyable_email('sender') && &notify_addr('sender')) {
	&debug("e_s: sending quarantine report via: $qmailinject to sender address ($returnpath)");
	print SM "To: $returnpath\n";
	$tmpsndrs = "$returnpath";
      } else {
	&debug("e_s: don't notify sender");
      }
    }elsif ($addr_type eq "psender") {
      if (!&is_unreplyable_email('sender') && &notify_addr('sender') && ($quarantine_event =~ /^(policy|perlscan)/i && $quarantine_event !~ /(gr[ae]ylist|virus)/i)) {
	&debug("e_s: sending policy quarantine report via: $qmailinject to psender address ($returnpath)");
	print SM "To: $returnpath\n";
	$tmpsndrs = "$returnpath";
      } else {
	&debug("e_s: don't notify psender");
      }
    } else {
      return;
    }
  } else {
    if (&notify_addr('admin') || (&notify_addr('nmladm') && !&is_unreplyable_email('sender')) || (&notify_addr('nmlvadm') && ($quarantine_event =~ /^(policy|perlscan)/i && $quarantine_event !~ /(gr[ae]ylist|virus)/i) && !&is_unreplyable_email('sender'))) {
      &debug("e_s: sending $polstring quarantine report via: $qmailinject to admin address ($QUARANTINE_CC)");
      print SM "To: $QUARANTINE_CC\n";
      $tmpsndrs .= "$QUARANTINE_CC";
    } else {
      &debug("e_s: don't notify admin");
    }
  }
  $tmpsubj="$destring found in sent message \"$headers{'subject'}\"";
  $tmpsubj =~ s/(\r|\0|\n)/ /g;
  print SM "Subject: $tmpsubj\n";
  print SM "Message-ID: <".&uniq_id."\@$hostname>\n";
  print SM "Auto-Submitted: auto-replied\n";
  if ($headers{'message-id'} ne "") {
    print SM "In-Reply-To: ",$headers{'message-id'},"\n";
    print SM "References: ",$headers{'message-id'},"\n";
  }
  print SM "MIME-Version: 1.0\n";
  print SM "Content-type: text/plain\n";
  print SM "Content-Transfer-Encoding: 8bit\n";
  #Only add these headers for Internet-incoming
  if ( $descriptive_hdrs && !$QS_RELAYCLIENT) {
    print SM "${V_HEADER}-Mail-From: $returnpath via $hostname\n";
    print SM "${V_HEADER}-Rcpt-To: $recips\n" if ($descriptive_hdrs eq "2");
    print SM "$V_HEADER: $VERSION ($SCANINFO $destring Found. \n";
    print SM " Processed in ",tv_interval($start_time,[gettimeofday])," secs)\n";
  }
  print SM "\n";
  if (&is_unreplyable_email('sender')) {
    print SM "
Attention: $V_FROMNAME.\n";
    print SM "
[This warning message is *not* being sent to the apparent originator
of the original message.  This address appears to be that of a
mailing list or other automated email system.]\n";
    print SM "\n---------------------------------------\n\n";
  } else {
    print SM "
Attention: $returnpath\n";
  }
  print SM "\n
A $destring was found in an Email message you sent. 
This Email scanner intercepted it and stopped the entire message
reaching its destination. 

The $destring was reported to be: 

$quarantine_description\n";
  if ($destring eq "virus") {
    print SM "\n
Please update your virus scanner or contact your IT support 
personnel as soon as possible as you may have a virus on your system.\n";
  } else {
    print SM "\n
Please contact your IT support personnel with any queries regarding this 
policy.\n";
  }
  print SM "\n
Your message was sent with the following envelope:

MAIL FROM: $returnpath
RCPT TO:   $recips 

... and with the following headers:\n\n";
  print SM "---\n";
  print SM "MAILFROM: $headers{'MAILFROM'}\n";
  print SM "RCPTTO: $headers{'RCPTTO'}\n";
  print SM "IP-Addr: $headers{'REMOTEIPADDR'}\n";
  print SM "$HEADERS\n";
  print SM "---\n";
  if ($addr_type !~ /sender/ ) {
    print SM "\n

The original message is kept in:

  $hostname:$scandir/quarantine/$vmaildir/new/$file_id

where the $V_FROMNAME can further diagnose it.

The Email scanner reported the following when it scanned that message:

--- 
$description
---\n";
  }
  close(SM);
  if ($log_details) {
    &log_msg("qmail-scanner","Clear:RC:1(127.0.0.1):",$elapsed_time,1100,$V_FROM,$tmpsndrs,$tmpsubj,$tmpmsgid,"quarantine-event.txt:1000");
  }
}

sub email_recips {
  my($recip)=@_;
  return if ($recip eq "");
  #Don't notify precips if this is NOT a "Policy block"
  if (&notify_addr('precips')) {
    return if ($quarantine_event !~ /^(policy|perlscan)/i);
  } else {
    #From now on precips is the same as recips
    $NOTIFY_ADDRS=~s/precips/recips/;
  }
  return if (!&notify_addr('recips'));
  my($HDR,$hdr,$tmprecips,$tmpsubj)='';
  my($tmpmsgid)= &uniq_id() . "-" . $V_FROM;

  open(SM,"|$qmailinject -h -f ''")||&error_condition("cannot open $qmailinject for sending quarantine report - $!");
  print SM "From: \"$V_FROMNAME\" <$V_FROM>\n";
  if (!&is_unreplyable_email('recips')) {
    &debug("e_r: sending quarantine report via: $qmailinject to recip address ($recip)");
    print SM "To: $recip\n";
  }
  $tmpsubj= "$destring found in received message \"$headers{'subject'}\"";
  $tmpsubj =~ s/(\r|\0|\n)/ /g;
  print SM "Subject: $tmpsubj\n";
  print SM "Message-ID: <".&uniq_id."\@$hostname>\n";
  if ($headers{'message-id'} ne "") {
    print SM "In-Reply-To: ",$headers{'message-id'},"\n";
    print SM "References: ",$headers{'message-id'},"\n";
  }
  print SM "Auto-Submitted: auto-replied\n";
  print SM "MIME-Version: 1.0\n";
  print SM "Content-type: text/plain\n";
  if ( $descriptive_hdrs ) {
    print SM "${V_HEADER}-Mail-From: $returnpath via $hostname\n";
    print SM "${V_HEADER}-Rcpt-To: $recip\n" if ($descriptive_hdrs eq "2");
    print SM "$V_HEADER: $VERSION ($SCANINFO $destring Found. \n";
    print SM " Processed in ",tv_interval($start_time,[gettimeofday])," secs)\n";
  }
  print SM "\n";
  print SM "
Attention: $recip\n";
  if (!&is_unreplyable_email('recips')) {
    if (&notify_addr('sender')) {
      print SM "
[A message has been sent to the originator, stating there is a virus
in the Email they just sent to you. No further action is required on
your part.]\n";
    }
  } else {
    print SM "

[This message was _not_ sent to the originator address, as that appears to
be a mailing-list or some other automated Email message]\n";
  }
  print SM "\nA $destring was found in an Email message sent to you. 
This Email scanner intercepted it and stopped the entire message
before it reached you. No further action is required on your part.\n";
  print SM "\nThe $destring was reported to be: 

$quarantine_description

Please contact your IT support personnel with any queries regarding this 
policy.

The message sent to you had the following envelope:

MAIL FROM: $returnpath
RCPT TO:   $recips 

... and with the following headers:\n\n";
  print SM "---\n";
  print SM "MAILFROM: $headers{'MAILFROM'}\n";
  print SM "RCPTTO: $headers{'RCPTTO'}\n";
  print SM "IP-Addr: $headers{'REMOTEIPADDR'}\n";
  print SM "$HEADERS\n";
  print SM "---\n";
  #print SM "\nLxOCALE_recips_quarantine\n";
  close(SM);
  if ($log_details) {
    &log_msg("qmail-scanner","Clear:$tag_score",$elapsed_time,1100,$V_FROM,$recip,$tmpsubj,$tmpmsgid,"quarantine-event.txt:1000");
  }
}

sub notify_addr {
  my($addr_type)=@_;
  #&debug("n_a: notify_addr (set to $NOTIFY_ADDRS) called with $addr_type");
  if (($NOTIFY_ADDRS =~ /$addr_type/ || $NOTIFY_ADDRS =~ /all/) && ($NOTIFY_ADDRS !~ /none/)) {
    return 1;
  } else {
    return 0;
  }
}

sub unzip_file {
  my($zipfile)=@_;
  my ($MAYBEZIP,$ztmp,$zfile,$zline,$zsize,$zip_status);
  
  &debug("u_f: potential zip archive file found ($zipfile).");
  &debug ("u_f: it is possibly a zip file, run unzip $unzip_options -t $ENV{'TMPDIR'}/$zipfile");
  $MAYBEZIP=`$unzip_binary $unzip_options -t $ENV{'TMPDIR'}/$zipfile 2>&1`;
  $zip_status=($? >> 8);
  
  if ( ($zip_status > 0) && ($zip_status !~ /^(1|2|3|51|81|82)$/) && ($MAYBEZIP !~ /skipping: /) ) {
    &debug("u_f: not a recognisable zip file ($MAYBEZIP)");
  } else {
    &debug ("u_f: it is a zip file");
    if ($MAYBEZIP =~ /skipping:.*password/) {
      &debug ("u_f: it is a password-protected zip file");
      $CRYPTO_TYPE="CR:ZIP(encrypted)";
    }
    if ($force_unzip) {
      &debug ("u_f: check size of contents before unzipping to disk");
      my $CHECK_ZIP_SIZE=`$unzip_binary $unzip_options -lv $ENV{'TMPDIR'}/$zipfile 2>&1`;
      open(ZIPPED,"$unzip_binary $unzip_options -lv  $ENV{'TMPDIR'}/$zipfile 2>&1|")||&error_condition("u_f: cannot open $ENV{'TMPDIR'}/$zipfile - $!");
      my $zip_file_size=0;
      while (<ZIPPED>) {
	$zip_file_size=$1 if (/^\s+([0-9]+)\s+/);
      }
      close ZIPPED ;
      &debug("u_f: this zip file unpacks to $zip_file_size bytes of content");
      if ($max_zip_size > 0 && $max_zip_size < $zip_file_size) {
	$quarantine_description="Disallowed zip file ($zipfile) - content exceeds maximum allowed size";
	&debug("u_f: $quarantine_description");
	$destring='problem';
	$quarantine_event="Policy:Oversized_ZIP";
	$quarantine_DOS=$quarantine_event;
	$description .= "\n---perlscanner results ---\n$destring '$quarantine_description' found in file $ENV{'TMPDIR'}/$zipfile";
	$file_desc .= "oversized_zip:$msg_size\t";
	return;
      }
      &debug("u_f: run $unzip_binary $unzip_options $ENV{'TMPDIR'}/$zipfile 2>&1");
      open(ZIPPED,"$unzip_binary $unzip_options   $ENV{'TMPDIR'}/$zipfile 2>&1|")||&error_condition("u_f: cannot open $ENV{'TMPDIR'}/$zipfile - $!");
      while (<ZIPPED>) {
	if (/^\s+\w+:\s+(.*)$/) {
	  ($ztmp=$1)=~s/^.*\///g;
	  #Grrr, I don't know if this'll be exploited, but I have to remove the whitespace...
	  #$ztmp=~s/\s+$//g;
	  #if ($ztmp ne "" && !grep(/^${ztmp}$/,@zipfile_list)) {
	  #&debug("u_f: adding file \"$ztmp\" to list of zipped files");
	  #push(@zipfile_list, $ztmp);
	  #}
	}
	if (/^\s+skipping:\s(.*)\s+(shrink|encrypted|incorrect password)/) {
	  $passwd_protected_zip++ if (!/^\s+skipping:\s(.*)\s+shrink/);
	  #grab these protected filenames for reports anyway.
	  $zfile = $1;
	  $zfile =~ s/^.*\///g;
	  $zfile =~ s/(^\s+|\s+$)//g;
	  #$file_desc .= "$zfile:$zsize\t";
	}
	close(ZIPPED);
	$zip_status=($? >> 8);
	if ($zip_status > 0 && ($zip_status !~ /^(1|2|3|51|81|82)$/ && !$passwd_protected_zip)) {
	  &error_condition("u_f: cannot close unzip (error code: $zip_status,$passwd_protected_zip) - $!");
	}
      }
    }
    #Only delete original zip file if it happily unpacked.
    if ( $zip_status eq 0 && -f "$ENV{'TMPDIR'}/$zipfile") {
	#system $rm_binary,"-f","$ENV{'TMPDIR'}/$zipfile";
	&debug("u_f: $zip_status, and successfully unzipped");
	#It may have been deleted, but you still want to see if 
	#it matches the perlscanner DB...
	#$zipfile=tolower($zipfile);
	#push(@zipfile_list, $zipfile) if (!grep(/^$zipfile$/,@zipfile_list));
	my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$zsize,$atime,$mtime,$ctime,$blksize,$blocks);
	($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$zsize,$atime,$mtime,$ctime,$blksize,$blocks) = stat("$zipfile");
	$file_desc .= "$zipfile:$zsize\t";
    } 
  }
}

sub deltatime {
  my ($delta,$current_time);
  $current_time = [gettimeofday];
  $delta =  tv_interval ($last_time, $current_time);
  $last_time=$current_time;
  return $delta;
}

sub qmail_parent_check {
  my $ppid=getppid;
  #&debug("q_s_c: PPID=$ppid");
  if ($ppid == 1)  {
    &debug("q_s_c: Whoa! parent process is dead! (ppid=$ppid) Better die too...");
    close(LOG);
    &cleanup;
    #Exit with temp error anyway - just to be real anal...
    exit 111; 
  }
}

sub check_and_grab_attachments {
  #This subroutine find attachments (e.g., MIME/binhex,uuencode) within e-mails

  if (/^MIME-Version:/i || ($headers{'content-type'} ne "" && $headers{'content-type'} !~ /^text\/plain/i)) {
    $indicates_attachments++;
    &debug("c_a_g: found MIME attachment") if ($indicates_attachments == 2);
  }
  #This finds MIME messages banged onto the bottom of bounces
  if (/^\-+ (Below this line|This) is a copy of the message/) {
    $indicates_attachments += 2;
    &debug("c_a_g: found hidden MIME attachment") if ($indicates_attachments == 2);
  }
  #This will define any text mail that contains a URL as requiring scanning - otherwise
  #some phishing attacks will geet past
  if ($indicates_attachments < 2 && /http:\/\/|www\.|[a-z0-9\-]+\.[a-z0-9\-]+\//i) {
    $indicates_attachments += 2;
    &debug("c_a_g: found URL in message - maybe phishy - better scan it");
  }
  #This finds BinHex attachments
  if (/^\(This file must be converted with BinHex/) {
    $indicates_attachments += 2;
    &debug("c_a_g: found hidden BinHex attachment") if ($indicates_attachments == 2);
  }
  my ($begin,$perms,$uufile,$uuextension,$uulength,$uuencoded_attachments,$begin_content);
  if (/^(begin) ([0-9][0-9][0-9]) (.*)\n$/) {
    &debug("Ooohhhh, a uuencoded attachment!");
    #Better reset this message back to potentially having attachments
    $plain_text_msg=0;
    $uuencoded_attachments++;
    $begin=$1;
    $perms=$2;
    $uufile=tolower($3);
    push(@uufile_list, $uufile) if(!grep(/^$uufile$/,@uufile_list));
    $uufile=~s/[^0-9a-z\_\-\.]/_/gi;
    $uufile =~ /\.([^\.]+)$/;
    $uuextension=$1;
    #Ensure the file extension isn't too long either...
    $uuextension=substr($uuextension,0,20); 
    $uulength=length($uufile);
    #Ensure the filelength isn't too large!
    if ( $uulength > $MAX_FILE_LENGTH) {
      &debug("uudecode output: gah! filename is > $MAX_FILE_LENGTH (actually $uulength), chopping...");
      $uufile=substr($uufile,0,$MAX_FILE_LENGTH).".".$uuextension;
    }
    return if (!$uudecode_binary);
    if (! -f "$ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile.uue") {
      open(UUIN,">$ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile.uue")||&error_condition("cannot open $ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile.uue - $!");
    } else {
      &error_condition("$ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile.uue already exists! - $!");
    }
    print UUIN "$begin 640 $uufile\n";
    print TMPFILE;
    $begin_content = "uuencode";
    while (<STDIN>) {
      if ($begin_content eq "uuencode" && $_ =~ /^(M35JJ|M35J0|M35KU|M35H\\|M35HN)/i) {
	&debug("w_c: looks like a Windows executable, filename=$uufile");
      }
      print UUIN;
      print TMPFILE;
      if (/^end/) {
	close(UUIN)||&error_condition("cannot close $ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile - $!");
	#uudecode it - toss away the error code - who cares if it's broken...
	&debug("c_a_g_u: $uudecode_binary $uudecode_pipe $ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile.uue");
	if (! -f  "$ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile") {
	  system("$uudecode_binary -o $ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile $ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile.uue 2>/dev/null");
	  rename("$ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile","$ENV{'TMPDIR'}/$uufile") if (!-f "$ENV{'TMPDIR'}/$uufile");
	} else {
	  &error_condition("$ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile already exists! - $!");
	}
	#open(UUOUT,"$uudecode_binary $uudecode_pipe $ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile.uue|");
	#open(UUFILE,">$ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile");
	#while (<UUOUT>) {
	  #print UUFILE;
	#}
	#close UUOUT;
	#close UUFILE;
	&debug("deleting uuencoded file as we have a decoded version of it now");
	unlink("$ENV{'TMPDIR'}/$file_id-$uuencoded_attachments-$uufile.uue") if ($DEBUG < 100);
	last;
      }
    }
  }
}

sub log_msg {
  my($msgtype,$status,$elapsed_time,$msgsize,$frm,$recips,$subj,$msgid,$attachs)=@_;
  my ($msg,$file);


  my $syslogtype='mail|info';

  if ($log_details eq "syslog") {

    $msgtype =~ s/\s/_/g;
    $msgtype .= "[$$]";
    $status =~ s/\s//g;
    $syslogtype='mail|warning' if ($status !~ /^Clear/);
    $elapsed_time =~ s/\s//g;
    $elapsed_time=0.0 if (!$elapsed_time);
    $elapsed_time=substr($elapsed_time,0,8);
    $frm =~ s/\s/_/g;
    $frm='<>' if (!$frm);
    $frm=substr($frm,0,100);
    $recips =~ s/\s/\|/g;
    $recips='<>' if (!$recips);
    $recips=substr($recips,0,100);
    $subj =~ s/\s/_/g;
    $subj='<>' if (!$subj);
    $subj=substr($subj,0,80);
    $msgid =~ s/\s/_/g;
    $msgid = '<>' if (!$msgid);
    $msgid=substr($msgid,0,80);
    $msgsize =~ s/\s//g;
    $attachs =~ s/\s$//g;
    #Sub any spaces for underscores then swap tabs for spaces,
    #syslog doesn't like tabs, so spaces in filenames have to go...
    $attachs =~ s/\ /_/g;
    $attachs =~ s/\t/ /g;
    if (!$attachs) {
      foreach $file (@uufile_list,@zipfile_list,@attachment_list) {
	$file =~ s/\s/_/g;
	$attachs .= "$file ";
      }
      $attachs =~ s/\s+$//g;
    }
    $attachs="$file_id-unpacked:$msg_size" if (!$attachs);
    #$attachs=substr($attachs,0,100);
    $msg = "$status $elapsed_time $msgsize $frm $recips $subj $msgid $attachs";
    #Do final santity check and remove all low-end chars - like NULL
    #I have no idea how some older syslogs would react to such things...
    $msg =~s/[\x00-\x09]//g;
    $msg =~ s/%/%%/g;
    #Now ensure syslog record isn't larger than max syslog size of 1024 chars
    $msg=substr($msg,0,1024);
    eval {
      $SIG{ALRM} = sub { die "Maximum time writing to syslog exceeded. syslog is hung/broken." };
      alarm 10;
      eval {
	syslog($syslogtype,"$msgtype: $msg");
      };
      if ($@) {
	setlogsock('inet');
	syslog('mail|info',"$msgtype: $msg");
      }
    };
    #The message is delivered - so no temp failure here - you just have to lose the log entry...
    alarm 0;
  } else {
    #No error checking - inability to write a log report shouldn't
    #stop the mail getting through!

    $msgtype =~ s/\t/ /g;
    $status =~ s/\s//g;
    $elapsed_time =~ s/\s//g;
    $elapsed_time=0 if (!$elapsed_time);
    $frm =~ s/\t/ /g;
    $frm='<>' if (!$frm);
    $recips =~ s/\t/ /g;
    $recips='<>' if (!$recips);
    $subj =~ s/\t/ /g;
    $subj='<>' if (!$subj);
    $msgid =~ s/\t/ /g;
    $msgid = '<>' if (!$msgid);
    $msgsize =~ s/\s//g;
    $attachs =~ s/\s$//g;
    $attachs =~ s/\t/ /g;
    $attachs="$file_id-unpacked:$msg_size"  if (!$attachs);
    $msg = "$status\t$elapsed_time\t$msgsize\t$frm\t$recips\t$subj\t$msgid\t$attachs";
    my $nowtime = strftime("%a, %d %b %Y %H:%M:%S %Z", localtime(time));
    open LOGMSG, ">>$log_details";
    print LOGMSG "$nowtime\t$msg\n";
    close LOGMSG;
  }
  &debug("$msgtype: $msg");
}


sub normalize_string {
  my ($type,$string)=@_;
  return $string;
} ;

###############################
#
##  END of standard subroutines
##  Virus-scanner specific subroutines automatically added below by setup.sh
##
###############################

#########################
## END of scanner definitions
##
#########################
